import os
from _decimal import Decimal
from datetime import date, timedelta
from orjson import orjson
from typing import Dict, List, Set

import pandas as pd

from . import helper
from ._typing import (
    AaveBalance,
    SupplyInfo,
    BorrowInfo,
    Supply,
    Borrow,
    InterestRateMode,
    RiskParameter,
    SupplyKey,
    BorrowKey,
    supply_to_dataframe,
    borrow_to_dataframe,
    SupplyAction,
    WithdrawAction,
    BorrowAction,
    RepayAction,
    LiquidationAction,
    DictCache,
    AaveDescription,
    AaveMarketStatus,
)
from .core import AaveV3CoreLib
from .. import DemeterError, TokenInfo, MarketTypeEnum
from .._typing import DECIMAL_0, UnitDecimal, ChainType
from ..broker import Market, MarketInfo, write_func
from ..data import CacheManager
from ..utils import get_formatted_predefined, STYLE, get_formatted_from_dict, console_text
from ..utils.application import require, float_param_formatter, to_decimal

DEFAULT_DATA_PATH = "./data"


class AaveV3Market(Market):
    """
    | AaveV3Market is the simulator of aave v3, here you can simulate some transactions like supply/borrow etc. this class also tracks value change.
    | AaveV3Market corresponds to a pool on chain, and one chan has one pool.

    :param market_info: key of this market
    :type market_info: MarketInfo
    :param risk_parameters_path: path to load risk parameters csv file, those files can be downloaded form https://www.config.fyi. One csv file for each chain.
    :type risk_parameters_path: str
    :param tokens: Tokens which will be used in this market
    :type tokens: List[TokenInfo]
    :param data: pool data for back test. downloaded by demeter-fetch, it's recommended to set data with load_data() or set_token_data() function
    :type data: DataFrame
    :param data_path: path to load data
    :type data_path: str
    """

    def __init__(
        self,
        market_info: MarketInfo,
        risk_parameters_path: str,
        tokens: List[TokenInfo] = None,
        data: pd.DataFrame = None,
        data_path: str = DEFAULT_DATA_PATH,
    ):
        super().__init__(market_info=market_info, data_path=data_path, data=data)
        tokens = tokens if tokens is not None else []  # just to set an initial value
        self._supplies: Dict[SupplyKey, SupplyInfo] = {}
        self._borrows: Dict[BorrowKey, BorrowInfo] = {}

        self._risk_parameters: pd.DataFrame | Dict[str, RiskParameter] = helper.load_risk_parameter(
            risk_parameters_path
        )

        # caches, since amounts and values are calculated by base_amount, and they are commonly used, cache is introduced to speed up back test.
        # when base amount or price has changed, they must be reset.
        # Maybe I'm just bring trouble on oneself
        self._collaterals_amount_cache = DictCache()
        self._supplies_amount_cache = DictCache()
        self._supplies_cache = DictCache()
        self._borrows_amount_cache = DictCache()
        self._borrows_cache = DictCache()

        self._market_status: AaveMarketStatus = None
        self._tokens: Set[TokenInfo] = set()
        self.add_token(tokens)

    REQUIRED_DATA_COLUMN = [
        "liquidity_rate",
        "stable_borrow_rate",
        "variable_borrow_rate",
        "liquidity_index",
        "variable_borrow_index",
    ]

    def __str__(self):
        from demeter.utils import orjson_default

        return orjson.dumps(self.description, default=orjson_default).decode()

    @property
    def description(self):
        """
        Get a brief description of this market
        """
        return AaveDescription(type(self).__name__, self._market_info.name, len(self._supplies), len(self._borrows))

    @property
    def data(self) -> pd.DataFrame:
        """
        | Get data attribute. data have multiple column index.
        | Columns are organized by token, like this:

        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+
        |                   | WETH           |                   |                     |                 |                      |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+
        |block_timestamp    | liquidity_rate |stable_borrow_rate |variable_borrow_rate | liquidity_index | variable_borrow_index|
        +===================+================+===================+=====================+=================+======================+
        |2023-08-15 00:00:00|              0 |                 0 |                   0 |            1    |             1        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+
        |2023-08-15 00:01:00|              0 |                 0 |                   0 |        1.001    |         1.001        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+
        |2023-08-15 00:02:00|              0 |                 0 |                   0 |        1.002    |         1.002        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+
        |2023-08-15 00:03:00|              0 |                 0 |                   0 |        1.003    |         1.003        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+
        |2023-08-15 00:04:00|              0 |                 0 |                   0 |        1.004    |         1.004        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+



        | So if you want to read a element in this dataframe, such as liquidity_rate of weth, you should do: data.iloc[0]["WETH"]["liquidity_rate"],
        | and if you access a column, you can do data["WETH"]["liquidity_rate"]
        | If you append a new column to data, the new column will be

        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+
        |                   | WETH           |                   |                     |                 |                      |     new_column       |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+
        |block_timestamp    | liquidity_rate |stable_borrow_rate |variable_borrow_rate | liquidity_index | variable_borrow_index|                      |
        +===================+================+===================+=====================+=================+======================+======================+
        |2023-08-15 00:00:00|              0 |                 0 |                   0 |            1    |             1        |             1        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+
        |2023-08-15 00:01:00|              0 |                 0 |                   0 |        1.001    |         1.001        |             2        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+
        |2023-08-15 00:02:00|              0 |                 0 |                   0 |        1.002    |         1.002        |             3        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+
        |2023-08-15 00:03:00|              0 |                 0 |                   0 |        1.003    |         1.003        |             4        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+
        |2023-08-15 00:04:00|              0 |                 0 |                   0 |        1.004    |         1.004        |             5        |
        +-------------------+----------------+-------------------+---------------------+-----------------+----------------------+----------------------+

        This bring an odd thing, if you access a element in new_column by access row first. data.iloc[0]["new_column"] will return a series instead a item. so data.iloc[0]["new_column"][0] will work
        and if you access by column first, data["new_column"] will return a series, so data["new_column"].iloc[0] will work, no need to write extra [0]

        """
        return self._data

    @property
    def risk_parameters(self) -> pd.DataFrame:
        """
        Get risk parameters
        """
        return self._risk_parameters

    def set_token_data(self, token_info: TokenInfo, token_data: pd.DataFrame):
        """
        Set aave pool data of one token. Usually demeter-fetch will keep one csv file for each token.

        :param token_info: which token to set
        :type token_info: TokenInfo
        :param token_data: data
        :type token_data: DataFrame
        """
        if self._data is not None and token_info.name in self._data:
            raise DemeterError(f"{token_info.name} has already set to data")
        if isinstance(token_data, pd.DataFrame):
            token_data = token_data.map(to_decimal)
            token_data.columns = pd.MultiIndex.from_tuples([(token_info.name, c) for c in token_data.columns])
            self._data = pd.concat([self._data, token_data], axis="columns")
        else:
            raise ValueError()

    def load_data(self, chain: ChainType, token_info_list: List[TokenInfo], start_date: date, end_date: date):
        """
        Load data from folder set in data_path. Those data file should be downloaded by demeter, and meet name rule. [chain]-aave_v3-[token_contract_address]-[date].minute.csv

        :param chain: chain type
        :type chain: ChainType
        :param token_info_list: tokens to load
        :type token_info_list: List[TokenInfo]
        :param start_date: start day
        :type start_date: date
        :param end_date: end day, the end day will be included
        :type end_date: date
        """
        self.logger.info(f"start load files from {start_date} to {end_date}...")
        for token_info in token_info_list:

            cache_key = CacheManager.get_cache_key(
                self.market_info.type.name, start_date, end_date, chain.name, token_info.name
            )
            cache_df = CacheManager.load(cache_key)
            if cache_df is not None:
                self.set_token_data(token_info, cache_df)
                continue

            day = start_date
            df = pd.DataFrame()
            if token_info.address == "":
                raise DemeterError(f"address of {token_info.name} not set")
            while day <= end_date:
                path = os.path.join(
                    self.data_path,
                    f"{chain.name.lower()}-aave_v3-{token_info.address}-{day.strftime('%Y-%m-%d')}.minute.csv",
                )
                if not os.path.exists(path):
                    raise IOError(
                        f"resource file {path} not found, please download with demeter-fetch: https://github.com/zelos-alpha/demeter-fetch"
                    )
                csv_converters = {n: to_decimal for n in AaveV3Market.REQUIRED_DATA_COLUMN}
                day_df = pd.read_csv(
                    path,
                    converters=csv_converters,
                    index_col=0,
                    parse_dates=True,
                )

                df = pd.concat([df, day_df])
                day += timedelta(days=1)

            CacheManager.save(cache_key, df)
            self.set_token_data(token_info, df)
        self.logger.info("data has been prepared")

    @property
    def tokens(self) -> Set[TokenInfo]:
        """
        Get tokens in aave back test
        """
        return self._tokens

    @property
    def supplies_value(self) -> Dict[SupplyKey, Decimal]:
        """
        Get value of all supplied token. unit is usd

        :return: value of all tokens
        :rtype: Dict[SupplyKey, Decimal]
        """
        if self._supplies_amount_cache.empty:
            for k, v in self._supplies.items():
                self._supplies_amount_cache.set(
                    k,
                    AaveV3CoreLib.get_amount(v.base_amount, self._market_status.data[k.token.name].liquidity_index)
                    * self._price_status[k.token.name],
                )
        return self._supplies_amount_cache.value

    @property
    def total_supply_value(self) -> Decimal:
        """
        Get sum supply value in this pool, unit is usd
        """
        return Decimal(sum(self.supplies_value.values()))

    @property
    def collateral_value(self) -> Dict[SupplyKey, Decimal]:
        """
        Get value of all collateral token. unit is usd

        :return: value of all collaterals
        :rtype: Dict[SupplyKey, Decimal]
        """
        if self._collaterals_amount_cache.empty:
            for k, v in self._supplies.items():
                if v.collateral:
                    self._collaterals_amount_cache.set(k, self.supplies_value[k])
        return self._collaterals_amount_cache.value

    @property
    def total_collateral_value(self) -> Decimal:
        """
        Get sum supply value in this pool, unit is usd
        """
        return Decimal(sum(self.collateral_value.values()))

    @property
    def borrows_value(self) -> Dict[BorrowKey, Decimal]:
        """
        Get value of all borrowed token. unit is usd

        :return: value of all borrows
        :rtype: Dict[BorrowKey, Decimal]
        """
        if self._borrows_amount_cache.empty:
            for k, v in self._borrows.items():
                self._borrows_amount_cache.set(
                    k,
                    AaveV3CoreLib.get_amount(
                        v.base_amount, self._market_status.data[k.token.name].variable_borrow_index
                    )
                    * self._price_status[k.token.name],
                )
        return self._borrows_amount_cache.value

    @property
    def total_borrows_value(self) -> Decimal:
        """
        Get sum borrow value in this pool, unit is usd
        """
        return Decimal(sum(self.borrows_value.values()))

    @property
    def supplies(self) -> Dict[SupplyKey, Supply]:
        """
        Get amounts and values of all supplies

        :return: Dict of supply information
        :rtype: Dict[SupplyKey, Supply]
        """
        if self._supplies_cache.empty:
            for key in self._supplies.keys():
                self._supplies_cache.set(key, self.get_supply(supply_key=key))
        return self._supplies_cache.value

    @property
    def supply_keys(self) -> List[SupplyKey]:
        """
        Get all supply keys
        """
        return list(self._supplies.keys())

    @property
    def borrows(self) -> Dict[BorrowKey, Borrow]:
        """
        Get amounts and values of all borrows

        :return: Dict of borrow information
        :rtype: Dict[BorrowKey, Borrow]

        """
        if self._borrows_cache.empty:
            for key in self._borrows.keys():
                self._borrows_cache.set(key, self.get_borrow(key))
        return self._borrows_cache.value

    @property
    def borrow_keys(self) -> List[BorrowKey]:
        """
        Get all borrow keys
        """
        return list(self._borrows.keys())

    @property
    def market_status(self) -> AaveMarketStatus:
        """
        Get market status
        """
        return self._market_status

    def set_market_status(
        self,
        data: AaveMarketStatus,
        price: pd.Series,
    ):
        """
        Set up pool status of this moment, such as liquidity index, supply/borrow rate, and price

        :param data: market status
        :type data: AaveMarketStatus
        :param price: current price of tokens involved
        :type price: Series
        """
        # TODO : debug test case
        super().set_market_status(data, price)
        if data.data is None:
            data.data = self.data.loc[data.timestamp]
        self._market_status = data
        self._borrows_amount_cache.reset()
        self._supplies_amount_cache.reset()
        self._collaterals_amount_cache.reset()
        self._borrows_cache.reset()
        self._supplies_cache.reset()

    @property
    def liquidation_threshold(self) -> Decimal:
        """
        Get liquidation threshold
        """
        return AaveV3CoreLib.total_liquidation_threshold(self.collateral_value, self._risk_parameters)

    @property
    def current_ltv(self) -> Decimal:
        """
        Get current ltv, it's the max ltv of current user
        """
        return AaveV3CoreLib.current_ltv(self.collateral_value, self._risk_parameters)

    @property
    def health_factor(self) -> Decimal:
        """
        Get health factor
        """
        return AaveV3CoreLib.health_factor(self.collateral_value, self.borrows_value, self._risk_parameters)

    @property
    def supply_apy(self) -> Decimal:
        """
        Calculate apy of all supplies
        """
        rate_dict: Dict[TokenInfo, Decimal] = {}
        for k in self.supplies.keys():
            rate_dict[k.token] = self._market_status.data[k.token.name].liquidity_rate

        return AaveV3CoreLib.get_apy(self.supplies_value, rate_dict)

    @property
    def borrow_apy(self) -> Decimal:
        """
        Calculate apy of all borrows
        """
        rate_dict: Dict[TokenInfo, Decimal] = {}
        for k in self._borrows.keys():
            if k.interest_rate_mode == InterestRateMode.variable:
                rate_dict[k.token] = self._market_status.data[k.token.name].variable_borrow_rate
            else:
                rate_dict[k.token] = self._market_status.data[k.token.name].stable_borrow_rate

        return AaveV3CoreLib.get_apy(self.borrows_value, rate_dict)

    @property
    def total_apy(self) -> Decimal:
        """
        Calculate total apy based on borrow/supply apy and amount
        """
        total_supplies = self.total_supply_value
        total_borrows = self.total_borrows_value
        supply_apy = self.supply_apy
        borrow_apy = self.borrow_apy
        return AaveV3CoreLib.safe_div_zero(
            supply_apy * total_supplies - borrow_apy * total_borrows, total_supplies - total_borrows
        )

    def get_supply(self, supply_key: SupplyKey = None, token_info: TokenInfo = None) -> Supply:
        """
        Get details of supply position, include value, amount, apy etc.

        :params supply_key: supply key, you can query by supply_key or token_info
        :type supply_key: SupplyKey
        :param token_info: token info, you can query by supply_key or token_info
        :type token_info: TokenInfo
        :return: details of supply position
        :rtype: Supply
        """
        key, token_info = AaveV3Market.__get_supply_key(supply_key, token_info)
        supply_info = self._supplies[key]
        supply_value = Supply(
            token=token_info,
            base_amount=supply_info.base_amount,
            collateral=supply_info.collateral,
            amount=supply_info.base_amount * self._market_status.data[key.token.name].liquidity_index,
            apy=AaveV3CoreLib.rate_to_apy(self._market_status.data[key.token.name].liquidity_rate),
            value=self.supplies_value[key],
            begin_supply_index=supply_info.begin_supply_index,
        )
        return supply_value

    def get_borrow(self, borrow_key: BorrowKey) -> Borrow:
        """
        Get details of borrow position. include type, amount, value etc.

        :params borrow_key: borrow key
        :type borrow_key: BorrowKey
        :return: details of borrow position
        :rtype: Borrow
        """
        borrow_info = self._borrows[borrow_key]
        return Borrow(
            token=borrow_key.token,
            base_amount=borrow_info.base_amount,
            interest_rate_mode=borrow_key.interest_rate_mode,
            amount=borrow_info.base_amount * self._market_status.data[borrow_key.token.name].variable_borrow_index,
            apy=AaveV3CoreLib.rate_to_apy(
                self.market_status.data[borrow_key.token.name].variable_borrow_rate
                if borrow_key.interest_rate_mode == InterestRateMode.variable
                else self.market_status.data[borrow_key.token.name].stable_borrow_rate
            ),
            value=self.borrows_value[borrow_key],
            begin_borrow_index=borrow_info.begin_borrow_index,
        )

    def add_token(self, token_info: TokenInfo | List[TokenInfo]):
        """
        Add one or an array of token to aave back test.

        :param token_info: tokens to add
        :type token_info: TokenInfo | List[TokenInfo]
        """
        if not isinstance(token_info, list):
            token_info = [token_info]
        for t in token_info:
            self._tokens.add(t)

    def get_market_balance(self) -> AaveBalance:
        """
        Get position and their values invested in aave market. Note: price will be read from self.market_status.price

        :return: values of positions in aave
        :rtype: AaveBalance
        """
        # supplys = self.supplies
        # borrows = self.borrows
        rounding = Decimal("0.0001")
        total_supplies = self.total_supply_value.quantize(rounding)
        total_borrows = self.total_borrows_value.quantize(rounding)
        net_worth = total_supplies - total_borrows

        supply_apy = self.supply_apy.quantize(rounding)
        borrow_apy = self.borrow_apy.quantize(rounding)
        net_apy = AaveV3CoreLib.safe_div_zero(
            supply_apy * total_supplies - borrow_apy * total_borrows, total_supplies - total_borrows
        )

        return AaveBalance(
            net_value=net_worth,
            supplies_count=len(self._supplies),
            borrows_count=len(self._borrows),
            liquidation_threshold=AaveV3CoreLib.safe_rounding(self.liquidation_threshold, rounding),
            health_factor=AaveV3CoreLib.safe_rounding(self.health_factor, rounding),
            borrows_value=total_borrows,
            supplies_value=total_supplies,
            collaterals_value=self.total_collateral_value.quantize(rounding),
            current_ltv=AaveV3CoreLib.safe_rounding(self.current_ltv, rounding),
            supply_apy=supply_apy,
            borrow_apy=borrow_apy,
            net_apy=net_apy,
        )

    # region for subclass to override
    def check_market(self):
        """
        Check market tokens before back test
        """
        super().check_market()
        require(len(self.tokens) > 0, "should set tokens")
        for t in self.tokens:
            require(
                t.name in self.risk_parameters.index,
                f"According to risk_parameter, {t.name} is not supported in this chain. ",
            )
            for col in AaveV3Market.REQUIRED_DATA_COLUMN:
                require((t.name, col) in self.data.columns, f"{t.name}.{col} not found in data")

    def update(self):
        """
        Trigger update of this market
        """
        self._liquidate()

    def formatted_str(self):
        """
        Return a brief description of this market in pretty format. Used for print in console.
        """
        value = get_formatted_predefined(f"{self.market_info.name}({type(self).__name__})", STYLE["header3"]) + "\n"
        token_dict = {"tokens": ",".join([t.name for t in self._tokens])}
        value += get_formatted_from_dict(token_dict) + "\n"
        balance = self.get_market_balance()
        value += (
            get_formatted_from_dict(
                {
                    "net_value": console_text.format_value(balance.net_value),
                    "health_factor": console_text.format_value(balance.health_factor),
                    "borrow_balance": console_text.format_value(balance.borrows_value),
                    "supply_balance": console_text.format_value(balance.supplies_value),
                    "collateral_balance": console_text.format_value(balance.collaterals_value),
                    "supply_apy": console_text.format_value(balance.supply_apy),
                    "borrow_apy": console_text.format_value(balance.borrow_apy),
                    "net_apy": console_text.format_value(balance.net_apy),
                }
            )
            + "\n"
        )
        value += get_formatted_predefined("Supplies", STYLE["key"]) + "\n"
        supply_df = supply_to_dataframe(self.supplies)
        value += supply_df.to_string() + "\n" if len(supply_df.index) > 0 else "Empty DataFrame\n"
        value += get_formatted_predefined("Borrows", STYLE["key"]) + "\n"
        borrow_df = borrow_to_dataframe(self.borrows)
        value += borrow_df.to_string() + "\n" if len(borrow_df.index) > 0 else "Empty DataFrame\n"

        return value

    # endregion

    @write_func
    @float_param_formatter
    def supply(self, token_info: TokenInfo, amount: Decimal | float, collateral: bool = True) -> SupplyKey:
        """
        | Supply asset to aave pool
        | Note:
        | 1. some token are not allow to collateral, it's decided by risk parameter
        | 2. if append amount to existing supply, collateral can not change.

        :param token_info: which token to supply
        :type token_info: TokenInfo
        :param amount: amount to supply
        :type amount: Decimal | float
        :param collateral: collateral or not, default is true.
        :type collateral: bool
        :return: key of new supply
        :rtype: SupplyKey
        """
        if collateral:
            require(self._risk_parameters.loc[token_info.name].canCollateral, "Can not supplied as collateral")
        token_status = self._market_status.data[token_info.name]
        #  calc in pool value
        pool_amount = AaveV3CoreLib.get_base_amount(amount, token_status.liquidity_index)

        self.broker.subtract_from_balance(token_info, amount)

        key = SupplyKey(token_info)
        if key not in self._supplies:
            self._supplies[key] = SupplyInfo(
                base_amount=Decimal(0), collateral=collateral, begin_supply_index=token_status.liquidity_index
            )
        else:
            require(self._supplies[key].collateral == collateral, "Collateral different from existing supply")
        self._supplies[key].base_amount += pool_amount

        self._supplies_amount_cache.reset()
        self._supplies_cache.reset()
        self._collaterals_amount_cache.reset()

        self._record_action(
            SupplyAction(
                market=self.market_info,
                token=token_info.name,
                amount=UnitDecimal(amount, token_info.name),
                collateral=collateral,
                deposit_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(self._supplies[key].base_amount, token_status.liquidity_index),
                    token_info.name,
                ),
            )
        )

        return key

    @staticmethod
    def __get_supply_key(supply_key: SupplyKey = None, token_info: TokenInfo = None):
        if supply_key is not None:
            key = supply_key
            token_info = key.token
        elif token_info is not None:
            key = SupplyKey(token_info)
        else:
            raise DemeterError("supply_key or token should be specified")
        return key, token_info

    @write_func
    def change_collateral(
        self, collateral: bool, supply_key: SupplyKey = None, token_info: TokenInfo = None
    ) -> SupplyKey:
        """
        Change collateral type of supply position. Health factor will be checked to prevent liquidation

        :param collateral: new value
        :type collateral: bool
        :param supply_key: key of supply, you can set by supply_key or token_info
        :type supply_key: SupplyKey
        :param token_info: token to supply, you can set by supply_key or token_info
        :type token_info: TokenInfo
        :return: key of new supply
        :rtype: SupplyKey
        """
        key, token_info = AaveV3Market.__get_supply_key(supply_key, token_info)
        old_collateral = self._supplies[key].collateral
        if old_collateral == collateral:
            return key

        self._supplies[key].collateral = collateral
        self._collaterals_amount_cache.reset()

        if (not collateral) and self.health_factor < AaveV3CoreLib.HEALTH_FACTOR_LIQUIDATION_THRESHOLD:
            # revert
            self._supplies[SupplyKey(token_info)].collateral = old_collateral
            self._collaterals_amount_cache.reset()
            raise AssertionError("health factor lower than liquidation threshold")

        return key

    @write_func
    @float_param_formatter
    def withdraw(
        self,
        supply_key: SupplyKey = None,
        amount: Decimal | float = None,
        token_info: TokenInfo = None,
    ):
        """
        Withdraw supply from aave pool.

        :param supply_key: key of supply, you can set by supply_key or token_info
        :type supply_key: SupplyKey
        :param amount: amount to withdraw, if set to None, will withdraw max available amount
        :type amount: Decimal | float
        :param token_info: which token to withdraw. you can set by supply_key or token_info
        :type token_info: TokenInfo
        """
        key, token_info = AaveV3Market.__get_supply_key(supply_key, token_info)
        token_status = self._market_status.data[token_info.name]
        supply = self.get_supply(key)
        if amount is None:
            amount = supply.amount
        require(amount != 0, "invalid amount")
        require(amount <= supply.amount, "not enough available user balance")

        # try calc new health factor after withdraw. if health factor is low, raise a error
        if self._supplies[key].collateral:
            old_base_amount = self._supplies[key].base_amount
            self._supplies[key].base_amount -= AaveV3CoreLib.get_base_amount(
                amount, self._market_status.data[key.token.name].liquidity_index
            )
            self._supplies_amount_cache.reset()
            self._collaterals_amount_cache.reset()
            if self.health_factor < AaveV3CoreLib.HEALTH_FACTOR_LIQUIDATION_THRESHOLD:
                raise AssertionError("health factor lower than liquidation threshold")
            self._supplies[key].base_amount = old_base_amount

        final_base_amount = self.__sub_supply_amount(key, amount)
        self.broker.add_to_balance(token_info, amount)
        self._record_action(
            WithdrawAction(
                market=self.market_info,
                token=token_info.name,
                amount=UnitDecimal(amount, token_info.name),
                deposit_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(final_base_amount, token_status.liquidity_index), token_info.name
                ),
            )
        )

        pass

    def get_max_withdraw_amount(self, supply_key: SupplyKey = None, token_info: TokenInfo = None) -> Decimal:
        """
        Get max available withdraw amount

        :param supply_key: key of supply to query, you can set by supply_key or token_info
        :type supply_key: SupplyKey
        :param token_info: which token to query. you can set by supply_key or token_info
        :type token_info: TokenInfo
        :return: max withdraw amount
        :rtype: Decimal
        """
        key, token_info = AaveV3Market.__get_supply_key(supply_key, token_info)
        return self.supplies[key].amount - AaveV3CoreLib.get_min_withdraw_kept_amount(
            token_info,
            self.collateral_value,
            self.borrows_value,
            self._risk_parameters,
            self._price_status[token_info.name],
        )

    def get_max_borrow_amount(self, token_info: TokenInfo) -> Decimal:
        """
        Get max token amount to borrow.

        :param token_info: token to query
        :param token_info: TokenInfo
        :return: max borrow amount
        :rtype: Decimal
        """
        value = AaveV3CoreLib.get_max_borrow_value(self.collateral_value, self.borrows_value, self.risk_parameters)
        return value / self._price_status[token_info.name]

    @write_func
    @float_param_formatter
    def borrow(
        self,
        token_info: TokenInfo,
        amount: Decimal | float = None,
        interest_rate_mode: InterestRateMode = InterestRateMode.variable,
    ) -> BorrowKey:
        """
        | Borrow token from aave pool.
        | Note:
        | 1. Risk parameter decides Which token is allowed to borrow
        | 2. If this borrow transaction will cause health factor below 1, an exception will be raised.
        | 3. If Borrow on existing borrow key, amount will be added to existing key

        :param token_info: Token to borrow
        :type token_info: TokenInfo
        :param amount: amount to borrow. if set to None, will borrow max amount.
        :type amount: Decimal | float
        :param interest_rate_mode: interest rate mode, default is variable.
        :type interest_rate_mode: InterestRateMode
        :return: key of new Borrows
        :rtype: BorrowKey
        """
        key = BorrowKey(token_info, interest_rate_mode)
        if amount is None:
            amount = self.get_max_borrow_amount(token_info)
        # check
        token_status = self._market_status.data[token_info.name]

        require(self._risk_parameters.loc[token_info.name, "canBorrow"], f"borrow is not enabled for {token_info.name}")
        collateral_balance = sum(self.collateral_value.values())
        require(collateral_balance != 0, "collateral balance is zero")
        current_ltv = self.current_ltv
        require(current_ltv != 0, "ltv validation failed")

        require(
            self.health_factor > AaveV3CoreLib.HEALTH_FACTOR_LIQUIDATION_THRESHOLD,
            "health factor lower than liquidation threshold",
        )

        value = amount * self._price_status.loc[token_info.name]
        collateral_needed = sum(self.borrows.values()) + value / current_ltv
        require(collateral_needed <= collateral_balance, "collateral cannot cover new borrow")

        if interest_rate_mode == InterestRateMode.stable:
            require(self._risk_parameters.loc[token_info.name, "canBorrowStable"], "stable borrowing not enabled")

            is_using_as_collateral = (token_info in self._supplies) and (
                self._supplies[SupplyKey(token_info)].collateral is True
            )

            require(
                (not is_using_as_collateral)
                or self._risk_parameters.loc[token_info.name, "LTV"] == 0
                or amount > self.broker.get_token_balance(token_info),
                "collateral same as borrowing currency",
            )
            # ignore pool amount check, I don't have pool amount

        # do borrow
        base_amount = AaveV3CoreLib.get_base_amount(amount, token_status.variable_borrow_index)

        if key not in self._borrows:
            self._borrows[key] = BorrowInfo(DECIMAL_0, token_status.variable_borrow_index)
        self._borrows[key].base_amount += base_amount

        self.broker.add_to_balance(token_info, amount)

        self._borrows_amount_cache.reset()
        self._borrows_cache.reset()

        self._record_action(
            BorrowAction(
                market=self.market_info,
                token=token_info.name,
                amount=UnitDecimal(amount, token_info.name),
                interest_rate_mode=interest_rate_mode,
                debt_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(self._borrows[key].base_amount, token_status.variable_borrow_index),
                    token_info.name,
                ),
            )
        )

        return key

    def get_max_repay_amount(
        self, key: BorrowKey = None, token_info: TokenInfo = None, interest_rate_mode: InterestRateMode = None
    ) -> Decimal:
        """
        Get max token amount to repay.

        :param key: key for borrow. Either fill key parameter or token_info+interest_rate_mode parameter.
        :type key: BorrowKey
        :param token_info: token to borrow, Either fill key parameter or token_info+interest_rate_mode parameter.
        :type token_info: TokenInfo
        :param interest_rate_mode: interest rate mode, Either fill key parameter or token_info+interest_rate_mode parameter.
        :type interest_rate_mode: InterestRateMode
        :return: max amount to repay
        :rtype: Decimal
        """
        (key, _, _) = AaveV3Market.__get_borrow_key(key, token_info, interest_rate_mode)
        return AaveV3CoreLib.get_amount(
            self._borrows[key].base_amount, self.market_status.data[key.token.name].variable_borrow_index
        )

    @staticmethod
    def __get_borrow_key(
        key: BorrowKey = None, token_info: TokenInfo = None, interest_rate_mode: InterestRateMode = None
    ):
        if key is None:
            if token_info is None:
                raise DemeterError("either key or token should be filled")
            key = BorrowKey(token_info, interest_rate_mode)
        token_info = key.token
        interest_rate_mode = key.interest_rate_mode
        return key, token_info, interest_rate_mode

    def _get_swap_amount(self, from_token: TokenInfo, to_token: TokenInfo, amount: Decimal, swap_fee=0):
        return amount * (1 - swap_fee) * self._price_status.loc[from_token.name] / self._price_status.loc[to_token.name]

    @write_func
    @float_param_formatter
    def repay(
        self,
        key: BorrowKey = None,
        payback_amount: Decimal | float = None,
        borrow_token: TokenInfo = None,
        interest_rate_mode: InterestRateMode = None,
        repay_with_collateral: bool = False,
        repay_collateral_token: TokenInfo = None,
    ):
        """
        Repay borrow. You can repay with collateral token or cash.

        :param key: key for borrow. Either fill key parameter or token_info+interest_rate_mode parameter.
        :type key: BorrowKey
        :param borrow_token: token to borrow, Either fill key parameter or token_info+interest_rate_mode parameter.
        :type borrow_token: TokenInfo
        :param interest_rate_mode: interest rate mode, Either fill key parameter or token_info+interest_rate_mode parameter.
        :type interest_rate_mode: InterestRateMode
        :param payback_amount: amount to pay back, if leave to None, will use max repay amount
        :type payback_amount: Decimal | float
        :param repay_with_collateral: If set to True, will repay with collateral, else will repay with cash
        :type repay_with_collateral: bool
        :param repay_collateral_token: Which collateral token is used to repay the debt, if repay_with_collateral==False, this parameter will not work,
        :type repay_collateral_token: TokenInfo

        """
        (key, borrow_token, interest_rate_mode) = AaveV3Market.__get_borrow_key(key, borrow_token, interest_rate_mode)
        # because liqThereshold<1, so repay will collateral will increase health factor, so there is no need to check health factor
        token_status = self._market_status.data[borrow_token.name]
        borrow = self.get_borrow(key)

        if payback_amount is None:
            payback_amount = borrow.amount

        if repay_with_collateral:
            if repay_collateral_token is None:
                repay_collateral_token = borrow_token
            repay_collateral_key = SupplyKey(repay_collateral_token)
            require(repay_collateral_key in self.supplies.keys(), f"token {repay_collateral_token} is not in supply")
            require(
                self.supplies[repay_collateral_key].collateral, f"token {repay_collateral_token} is not in collateral"
            )

            required_amount_in_collateral_token = self._get_swap_amount(
                borrow_token, repay_collateral_token, payback_amount
            )
            supply = self.get_supply(supply_key=repay_collateral_key)
            if required_amount_in_collateral_token > supply.amount:
                # contract will change payback amount instead of raise a error
                payback_amount = self._get_swap_amount(repay_collateral_token, borrow_token, supply.amount)

        payback_base_amount = AaveV3CoreLib.get_base_amount(payback_amount, token_status.variable_borrow_index)

        require(payback_base_amount > 0, "invalid amount")
        require(self._borrows[key].base_amount > 0, "no debt of selected type")
        require(round(self._borrows[key].base_amount - payback_base_amount, 18) >= 0, "amount exceed debt")
        if repay_with_collateral:
            payback_amount_in_collateral = self._get_swap_amount(borrow_token, repay_collateral_token, payback_amount)
            self.__sub_supply_amount(SupplyKey(repay_collateral_token), payback_amount_in_collateral)
        else:
            self.broker.subtract_from_balance(borrow_token, payback_amount)
        debt = self.__sub_borrow_amount(key, payback_amount)
        self._record_action(
            RepayAction(
                market=self.market_info,
                token=borrow_token.name,
                amount=UnitDecimal(payback_amount, borrow_token.name),
                interest_rate_mode=interest_rate_mode,
                debt_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(debt, token_status.variable_borrow_index), borrow_token.name
                ),
            )
        )

        pass

    def __sub_supply_amount(self, key: SupplyKey, amount: Decimal) -> Decimal:
        if key not in self._supplies:
            if amount == Decimal(0):
                return Decimal(0)
            else:
                raise DemeterError(f"{key} not exist in supplies")
        self._supplies[key].base_amount = helper.sub_base_amount(
            self._supplies[key].base_amount,
            AaveV3CoreLib.get_base_amount(amount, self._market_status.data[key.token.name].liquidity_index),
        )
        if self._supplies[key].collateral:
            self._collaterals_amount_cache.reset()
        self._supplies_amount_cache.reset()
        self._supplies_cache.reset()
        if self._supplies[key].base_amount == DECIMAL_0:
            if self._supplies[key].collateral:
                self._collaterals_amount_cache.reset()
            del self._supplies[key]
            return DECIMAL_0
        else:
            return self._supplies[key].base_amount

    def __sub_borrow_amount(self, key: BorrowKey, amount: Decimal) -> Decimal:
        if key not in self._borrows:
            if amount == Decimal(0):
                return Decimal(0)
            else:
                raise DemeterError(f"{key} not exist in borrows")
        self._borrows[key].base_amount = helper.sub_base_amount(
            self._borrows[key].base_amount,
            AaveV3CoreLib.get_base_amount(amount, self._market_status.data[key.token.name].variable_borrow_index),
        )
        self._borrows_amount_cache.reset()
        self._borrows_cache.reset()
        if self._borrows[key].base_amount == DECIMAL_0:
            del self._borrows[key]
            return DECIMAL_0
        else:
            return self._borrows[key].base_amount

    @write_func
    def _liquidate(self):
        """
        | Do passive liquidate. If health factor is below 1, liquidate will be triggered. And user can not make a liquidation transaction. Because currently demeter doesn't support that.
        | First is to decide when to liquidate, if in current loop, health factor is below 1, liquidation will happen immediately. Triggered by update function.
        | Second step to choose the assets. Here we will pick the most valuable collateral asset and least valuable debt.
        | After liquidation, if health factor is still below 1. Will choose a pair of collateral and debt to liquidate again.
        | But if a debt asset has liquidated, It will not be liquidated again.
        """
        health_factor = self.health_factor
        has_liquidated: List[BorrowKey] = []
        # health_factor ==0 means there are no collateral to liquidate
        # health_factor > 0.000001 to avoid calculate error
        while 0 < health_factor < AaveV3CoreLib.HEALTH_FACTOR_LIQUIDATION_THRESHOLD:
            # choose which token and how much to liquidate

            # choose the smallest delt
            borrows = self.borrows
            supplys = self.supplies

            min_borrow_key = None
            min_borrow_value = Decimal(10e21)  # a very large number
            for k, v in borrows.items():
                if min_borrow_value >= v.value and (k not in has_liquidated):
                    min_borrow_value = v.value
                    min_borrow_key = k
            # choose the biggest collateral
            max_supply_key = None
            max_supply_value = Decimal(0)
            for k, v in supplys.items():
                if v.collateral and max_supply_value <= v.value:
                    max_supply_value = v.value
                    max_supply_key = k
            # if a token has liquidated, but health_factor still < 1, it should not be liquidated again.
            # because if 0.95 < health_factor < 1, only half of this token will be liquidated. if liquidate this token again, will only liquidate 1/4
            # so health_factor will never go above 1

            has_liquidated.append(min_borrow_key)

            try:
                self._do_liquidate(max_supply_key.token, min_borrow_key.token, min_borrow_value)
            except AssertionError:
                # if a liquidated is rejected, choose another delt token to liquidate
                pass
            health_factor = self.health_factor

    def _do_liquidate(self, collateral_token: TokenInfo, delt_token: TokenInfo, delt_value_to_cover: Decimal):
        """
        Make a liquidation transaction;

        :param collateral_token: Which collateral token will be used
        :type collateral_token: TokenInfo
        :param delt_token: Which debt token will be repaid
        :type delt_token: TokenInfo
        :param delt_value_to_cover: total delt token value to repay. unit is usd
        :type delt_value_to_cover: Decimal

        """
        old_health_factor = self.health_factor
        borrow_index = self._market_status.data[delt_token.name].variable_borrow_index
        supply_index = self._market_status.data[delt_token.name].liquidity_index

        stable_key = BorrowKey(delt_token, InterestRateMode.stable)
        variable_key = BorrowKey(delt_token, InterestRateMode.variable)
        collateral_key = SupplyKey(collateral_token)
        liquidation_bonus = self._risk_parameters.loc[collateral_token.name].liqBonus

        # _calculateDebt
        stable_delt = self.get_borrow(stable_key).amount if stable_key in self._borrows else DECIMAL_0
        variable_delt = self.get_borrow(variable_key).amount if variable_key in self._borrows else DECIMAL_0
        total_debt = stable_delt + variable_delt
        close_factor = (
            AaveV3CoreLib.DEFAULT_LIQUIDATION_CLOSE_FACTOR
            if old_health_factor > AaveV3CoreLib.CLOSE_FACTOR_HF_THRESHOLD
            else AaveV3CoreLib.MAX_LIQUIDATION_CLOSE_FACTOR
        )
        max_liquidatable_debt = total_debt * close_factor
        actual_debt_to_liquidate = (
            max_liquidatable_debt if delt_value_to_cover > max_liquidatable_debt else delt_value_to_cover
        )

        # validate delt
        is_collateral_enabled = (
            self._risk_parameters.loc[collateral_token.name].liqThereshold != 0
            and self._supplies[collateral_key].collateral
        )

        require(is_collateral_enabled, "collateral cannot be liquidated")
        require(total_debt != DECIMAL_0, "specified currency not borrowed by user")

        user_collateral_balance = (
            self._supplies[SupplyKey(collateral_token)].base_amount
            * self._market_status.data[collateral_token.name].liquidity_index
        )

        # calculate actual amount
        should_collateral = (
            self._price_status.loc[delt_token.name]
            * actual_debt_to_liquidate
            / self._price_status.loc[collateral_token.name]
        )
        max_collateral_to_liquidate = should_collateral * (1 + liquidation_bonus)

        if max_collateral_to_liquidate > user_collateral_balance:
            actual_collateral_to_liquidate = user_collateral_balance
            actual_debt_to_liquidate = (
                self._price_status.loc[collateral_token.name] * actual_collateral_to_liquidate
            ) / (self._price_status.loc[delt_token.name] * (1 + liquidation_bonus))
        else:
            actual_collateral_to_liquidate = max_collateral_to_liquidate
            actual_debt_to_liquidate = actual_debt_to_liquidate
        self._supplies[collateral_key].base_amount = helper.sub_base_amount(
            self._supplies[collateral_key].base_amount,
            AaveV3CoreLib.get_base_amount(actual_collateral_to_liquidate, supply_index),
        )

        if self._supplies[collateral_key].base_amount == 0:
            del self._supplies[collateral_key]
        if variable_delt > actual_debt_to_liquidate:
            vari_debt_remaining_base = self.__sub_borrow_amount(variable_key, actual_debt_to_liquidate)
            stable_debt_remaining_base = (
                self._borrows[stable_key].base_amount if stable_key in self._borrows else DECIMAL_0
            )
            vari_debt_liquidated = actual_debt_to_liquidate
            stable_debt_liquidated = DECIMAL_0

        else:
            vari_debt_liquidated = variable_delt
            stable_debt_liquidated = actual_debt_to_liquidate - variable_delt
            vari_debt_remaining_base = self.__sub_borrow_amount(variable_key, variable_delt)
            stable_debt_remaining_base = self.__sub_borrow_amount(stable_key, stable_debt_liquidated)

        self._borrows_amount_cache.reset()
        self._borrows_cache.reset()
        self._supplies_amount_cache.reset()
        self._supplies_cache.reset()
        self._collaterals_amount_cache.reset()

        self._record_action(
            LiquidationAction(
                market=self.market_info,
                collateral_token=collateral_token.name,
                debt_token=delt_token.name,
                delt_to_cover=UnitDecimal(delt_value_to_cover, delt_token.name),
                collateral_used=UnitDecimal(actual_collateral_to_liquidate, collateral_token.name),
                variable_delt_liquidated=UnitDecimal(vari_debt_liquidated, delt_token.name),
                stable_delt_liquidated=UnitDecimal(stable_debt_liquidated, delt_token.name),
                health_factor_before=old_health_factor,
                health_factor_after=self.health_factor,
                collateral_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(
                        self._supplies[collateral_key].base_amount if collateral_key in self._supplies else DECIMAL_0,
                        supply_index,
                    ),
                    collateral_token.name,
                ),
                variable_debt_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(vari_debt_remaining_base, borrow_index), delt_token.name
                ),
                stable_delt_after=UnitDecimal(
                    AaveV3CoreLib.get_amount(stable_debt_remaining_base, borrow_index), delt_token.name
                ),
            )
        )

    def _resample(self, freq: str):
        self._data = self.data.resample(freq).last()



import os
from datetime import datetime, timedelta, date
from typing import NamedTuple

import pandas as pd
import pickle
from dataclasses import dataclass

CACHE_PATH = os.path.join(os.path.expanduser("~"), ".demeter")
CACHE_CONFIG_PATH = os.path.join(CACHE_PATH, "config.pkl")
CACHE_KEEP_DAYS = 30


class CacheKey(NamedTuple):
    market: str
    start: str
    end: str
    chain: str = ""
    address: str = ""


@dataclass
class CacheItem:
    create_time: datetime
    last_visit: datetime
    file_name: str


class CacheManager:
    @staticmethod
    def get_cache_key(market: str, start: date, end: date, chain: str = "", address: str = ""):
        return CacheKey(market, start.strftime("%y%m%d"), end.strftime("%y%m%d"), chain, address)

    @staticmethod
    def prepare_cache():
        if not os.path.exists(CACHE_PATH):
            os.mkdir(CACHE_PATH)
        if os.path.exists(CACHE_CONFIG_PATH):

            with open(CACHE_CONFIG_PATH, "rb") as f:
                config = pickle.load(f)
            # remove old files
            key_to_remove = []
            for k, v in config.items():
                if v.last_visit + timedelta(days=CACHE_KEEP_DAYS) < datetime.now():
                    os.remove(os.path.join(CACHE_PATH, v.file_name))
                    key_to_remove.append(k)
            for k in key_to_remove:
                del config[k]
            if len(key_to_remove) > 0:
                with open(CACHE_CONFIG_PATH, "wb") as f:
                    pickle.dump(config, f)
        else:
            config = {}
            with open(CACHE_CONFIG_PATH, "wb") as f:
                pickle.dump(config, f)

    @staticmethod
    def save(key: CacheKey, df: pd.DataFrame):

        CacheManager.prepare_cache()
        with open(CACHE_CONFIG_PATH, "rb") as f:
            config = pickle.load(f)
        if key in config:
            return

        file_name = f"{key.market}_{key.chain}_{key.start}_{key.end}_{key.address}.feather"
        config[key] = CacheItem(create_time=datetime.now(), last_visit=datetime.now(), file_name=file_name)
        df.to_feather(os.path.join(CACHE_PATH, file_name), compression="lz4")

        with open(CACHE_CONFIG_PATH, "wb") as f:
            pickle.dump(config, f)

    @staticmethod
    def load(key: CacheKey) -> pd.DataFrame | None:
        if not os.path.exists(CACHE_CONFIG_PATH):
            return None
        with open(CACHE_CONFIG_PATH, "rb") as f:
            config = pickle.load(f)
        if key not in config:
            return None

        path = os.path.join(CACHE_PATH, config[key].file_name)
        if not os.path.exists(path):
            del config[key]
        else:
            config[key].last_visit = datetime.now()
        with open(CACHE_CONFIG_PATH, "wb") as f:
            pickle.dump(config, f)
        return pd.read_feather(path)



import logging

from datetime import datetime

import json

from dataclasses import dataclass

from typing import NamedTuple

from decimal import Decimal
from enum import Enum


class Formats:
    # follow the document here: https://python-reference.readthedocs.io/en/latest/docs/functions/format.html
    global_num_format: str = ".8g"


# constant value for number 1
DECIMAL_0 = Decimal(0)

# constant value for number 0
DECIMAL_1 = Decimal(1)


class TimeUnitEnum(Enum):
    """
    Time unit of moving average,

    * minute
    * hour
    * day
    """

    minute = 1
    hour = 60
    day = 60 * 24


class UnitDecimal(Decimal):
    """
    Decimal with unit, such a 1 eth.

    It's inherit from Decimal, but considering performance issues, calculate function is not override,
    so if you do calculate on this object, return type will be Decimal

    :param number: number to keep
    :type number: Decimal
    :param unit: unit of the number, e.g. eth
    :type unit: str
    """

    __integral = Decimal(1)

    def __new__(cls, value, unit: str = ""):
        obj = Decimal.__new__(cls, value)
        obj._unit = unit
        return obj

    def to_str(self):
        """
        Get formatted string like "12.34 eth". Decimal format is predefined by self.output_format attribute

        :return: formatted string
        :rtype: str
        """
        dec = self.quantize(DECIMAL_1) if (self == self.to_integral() and self < 1e29) else self.normalize()
        return "{:{}} {}".format(dec, Formats.global_num_format, self._unit)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = value


@dataclass
class TokenInfo:
    """
    Identity for a token, will be used as key for token dict.

    :param name: token symbol, will be set as unit of a token value, e.g. usdc
    :type name: str
    :param decimal: decimal of this token, e.g. 6
    :type decimal: int
    :param address: Address of token, for aave market, this attribute has to be filled to load data.
    :type decimal: str
    """

    name: str
    decimal: int
    address: str

    def __init__(self, name: str, decimal: int, address: str = ""):
        self.name = name.upper()
        self.decimal = decimal
        self.address = address.lower()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, TokenInfo):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return self.name.__hash__()


class DemeterError(RuntimeError):
    def __init__(self, message):
        self.message = message


class DemeterWarning(RuntimeWarning):
    def __init__(self, message):
        self.message = message


class ChainType(int, Enum):
    """
    Enum for chains
    """

    ethereum = 1
    polygon = 137
    optimism = 69
    arbitrum = 42161
    celo = 42220
    bsc = 56
    base = 8453
    avalanche = 43114
    fantom = 250
    harmony = 1666600000


@dataclass
class MarketDescription:
    type: str
    """market type string"""
    name: str
    """market name"""

@dataclass
class DemeterLog:
    time: datetime
    message: str
    level: int = logging.INFO


USD = TokenInfo("USD", 0)

STABLE_COINS = [
    "USD",
    "USDT",
    "USDC",
    "DAI",
    "FDUSD",
    "PYUSD",
    "TUSD",
    "EDLC",
    "USDE",
    "FRAX",
    "USDB",
    "USDY",
    "USDJ",
    "CRVUSD",
    "EURS",
    "USDP",
    "GUSD",
    "USDX",
    "LUSD",
    "GHO",

]



from .market import DeribitOptionMarket
from ._typing import (
    DeribitMarketStatus,
    OptionPosition,
    OptionKind,
    OptionMarketBalance,
    BuyAction,
    SellAction,
    ExpiredAction,
    DeliverAction,
    DERIBIT_OPTION_FREQ,
    DeribitOptionDescription,
    InsufficientBalanceError,
)
from .helper import round_decimal, decode_instrument



import math
from decimal import Decimal, getcontext
from typing import Tuple, NamedTuple

from .liquitidy_math import get_sqrt_ratio_at_tick, get_liquidity, get_amounts
from .. import DemeterError

Q96 = Decimal(2**96)
SQRT_1p0001 = math.sqrt(Decimal(1.0001))
getcontext().prec = 35  # default is 28, 33 is good enough for 3000
MIN_ERROR = Decimal("1e-31")


def _from_x96(number: int) -> Decimal:
    """
    decimal divide 2 ** 96

    :param number: sqrted x96 price
    :return: sqrted price
    """
    return Decimal(number) / Q96


def _to_x96(sqrt_price: Decimal) -> int:
    """
    decimal multiple 2 ** 96

    :param sqrt_price: price in base unit
    :return: result
    """
    if not isinstance(sqrt_price, Decimal):
        sqrt_price = Decimal(sqrt_price)
    return int(sqrt_price * Q96)


def sqrt_price_x96_to_base_unit_price(
    sqrt_price_x96, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool
) -> Decimal:
    """
    convert sqrt x96 price to decimal

    :param sqrt_price_x96: sqrt x96 price
    :param token_0_decimal: decimal of token 0
    :param token_1_decimal: decimal of token 1
    :param is_token0_quote: is token 0 the quote token
    :return: price in base unit, e.g. 1234.56 eth/usdc
    """
    sqrt_price = _from_x96(sqrt_price_x96)
    pool_price = sqrt_price**2 * Decimal(10 ** (token_0_decimal - token_1_decimal))

    return Decimal(1 / pool_price) if is_token0_quote else pool_price


def base_unit_price_to_sqrt_price_x96(
    price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool
) -> int:
    """
    convert quote price to sqrt

    :param price: price in base/quote
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_quote: token 0 is quote token
    :return: sqrt price x96
    """
    # quote price->add decimal pool price->sqrt_price ->ticker

    price = 1 / price if is_token0_quote else price
    atomic_unit_price = price / Decimal(10 ** (token_0_decimal - token_1_decimal))
    sqrt_price = Decimal.sqrt(atomic_unit_price)
    return _to_x96(sqrt_price)


# can round by spacing?
def sqrt_price_x96_to_tick(sqrt_price_x96: int) -> int:
    """
    convert sqrt_priceX96 to tick

    :param sqrt_price_x96: sqrt x96 price
    :return: tick price
    """
    sqrt_price = _from_x96(sqrt_price_x96)
    return _sqrt_price_to_tick(sqrt_price)


def _sqrt_price_to_tick(sqrt_price: Decimal) -> int:
    """
    pool price to tick price

    :param sqrt_price: sqrt price
    :return: tick price
    """
    return int(math.log(sqrt_price, SQRT_1p0001))


def tick_to_sqrt_price_x96(tick: int) -> int:
    """
    convert tick data to sqrt X96 price

    :param tick: tick price
    :return: sqrt x96 price
    """
    return get_sqrt_ratio_at_tick(tick)


def tick_to_base_unit_price(tick: int, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool):
    """
    get quote price from tick price

    :param tick: tick data
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_quote: quote on token0
    :return: quote price
    """
    sqrt_price_x96 = get_sqrt_ratio_at_tick(tick)
    atomic_unit_price = _from_x96(sqrt_price_x96) ** 2
    pool_price = atomic_unit_price * Decimal(10 ** (token_0_decimal - token_1_decimal))
    return Decimal(1 / pool_price) if is_token0_quote else pool_price


def base_unit_price_to_tick(price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool) -> int:
    """
    quote price to tick price

    :param price: price in base unit, e.g. 1234.56 eth/usdc
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_quote: token 0 is quote token
    :return: tick price
    """

    # quote price->add decimal pool price->sqrt_price ->tick
    price = 1 / price if is_token0_quote else price
    atomic_unit_price = price / Decimal(10 ** (token_0_decimal - token_1_decimal))
    sqrt_price = Decimal.sqrt(atomic_unit_price)
    return _sqrt_price_to_tick(sqrt_price)


def from_atomic_unit(atomic_unit_amount: int, decimal: int) -> Decimal:
    """
    Convert token amount to wei

    :param atomic_unit_amount: token amount
    :param decimal: decimal of token
    :return: token amount in base unit
    """
    return Decimal(int(atomic_unit_amount)) / Decimal(10**decimal)


class Greeks(NamedTuple):
    delta: Decimal
    gamma: Decimal


def get_greeks(P: Decimal, L: Decimal, H: Decimal) -> Greeks:
    # get greeks for 1 u

    if P >= H:
        return Greeks(Decimal(0), Decimal(0))

    liq = 1 / (2 - L.sqrt() - 1 / H.sqrt())

    if P < L:
        delta = liq * (1 / L.sqrt() - 1 / H.sqrt())
        return Greeks(delta, Decimal(0))

    delta = liq * (1 / P.sqrt() - 1 / H.sqrt())
    gamma = Decimal("-0.5") * liq * (P ** Decimal("-1.5"))

    return Greeks(delta, gamma)


def get_swap_value(swap_from_token_val, swap_to_token_val, fee_rate, final_ratio):
    # if needed value rate is K, calculate how many value to swap, need deduct swap fee
    # final_ratio = final_from_token_val / final_to_token_value

    # Vfrom + Vto = Vfrom_after  + Vto_after + Vswap * fee_rate
    # Vfrom - Vswap = Vfrom_after
    # Vto + Vswap = Vto_after + Vswap * fee_rate
    # Vfrom_after / Vto_after = final_ratio

    # known value: Vfrom, Vto, fee_rate, final_ratio
    # unknow value: Vfrom_after, Vto_after
    # wanted value: Vswap
    return (swap_from_token_val - swap_to_token_val * final_ratio) / (final_ratio - final_ratio * fee_rate + 1)


def get_swap_value_with_part_balance_used(
    swap_from_token_val, swap_to_token_val, total_val_after, fee_rate, final_ratio
):
    # Given current balance of A and B, and suppose you want to invest total_val_after in lp,
    # Ratio of a and b after swap is known
    # This function will calculate how much to swap,

    # Vfrom + Vto = V
    # Vswap * fee_rate + Vfrom_after + Vto_after = total_val_after
    # Vfrom_after / Vto_after = final_ratio
    # Vto + Vswap * (1 - fee_rate) = Vto_after

    if total_val_after > swap_from_token_val + swap_to_token_val:
        raise DemeterError("Target value exceed your balance")
    swap_value = (total_val_after - final_ratio * swap_to_token_val - swap_to_token_val) / (
        final_ratio - final_ratio * fee_rate + 1
    )
    total_val_after_no_fee = total_val_after - swap_value * fee_rate
    swap_to_token_val_after = total_val_after_no_fee / (final_ratio + 1)
    return total_val_after_no_fee - swap_to_token_val_after, swap_to_token_val_after, swap_value


def nearest_usable_tick(tick: int, tick_spacing: int):
    """
    Returns the closest tick that is nearest a given tick and usable for the given tick spacing
    copied from: https://github.com/Uniswap/v3-sdk/blob/main/src/utils/nearestUsableTick.ts
    :param tick: the target tick
    :param tick_spacing: the spacing of the pool
    :return:
    """
    rounded = round(tick / tick_spacing) * tick_spacing
    if rounded < -887272:
        return rounded + tick_spacing
    elif rounded > 887272:
        return rounded - tick_spacing
    else:
        return rounded


class TickResult(NamedTuple):
    final_price: Decimal
    rate: Decimal
    center_tick: int
    upper: int
    lower: int
    upper_delta: int
    lower_delta: int


def find_tick_range_at_rate(
    price: Decimal,
    rate: Decimal,
    tick_spacing: int,
    decimal0: int,
    decimal1: int,
    is_0_quote: bool,
    error=Decimal("0.00001"),
):
    """
    Find iteratively, for a specific price, what tick range can make the quantities of two tokens exactly equal to a specific ratio.
    :param price: center price, it will be trimed according to tick space
    :param rate: the rate you want. amount1/amount0
    :param tick_spacing: tick spacing of the pool
    :param decimal0: decimal 0
    :param decimal1: decimal 1
    :param is_0_quote: token 0 is the quote token
    :param error: error of rate
    :return: An object contains new price, actual rate, and tick range, if can not find a proper tick range, return none, you can make error larger
    """
    center_tick = base_unit_price_to_tick(price, decimal0, decimal1, is_0_quote)
    center_tick = nearest_usable_tick(center_tick, tick_spacing)
    idx = tick_spacing

    sqrt_price = tick_to_sqrt_price_x96(center_tick)
    price = sqrt_price_x96_to_base_unit_price(sqrt_price, decimal0, decimal1, is_0_quote)
    if is_0_quote:
        token0_amount, token1_amount = price, Decimal(1)
    else:
        token1_amount, token0_amount = price, Decimal(1)

    rate = rate.quantize(error)

    while center_tick + idx <= 887272:
        print("trying", center_tick + idx)
        upper = center_tick + idx
        lower_idx = idx
        val1, val0 = 0, 1
        while val1 / val0 <= rate:
            lower_idx += tick_spacing
            lower = center_tick - lower_idx
            liq = get_liquidity(sqrt_price, lower, upper, token0_amount, token1_amount, decimal0, decimal1)
            amount0, amount1 = get_amounts(sqrt_price, lower, upper, liq, decimal0, decimal1)
            if is_0_quote:
                val0, val1 = amount0, amount1 * price
            else:
                val0, val1 = amount0 * price, amount1

            if (val1 / val0).quantize(error) == rate:
                return TickResult(
                    final_price=tick_to_base_unit_price(center_tick, decimal0, decimal1, is_0_quote),
                    rate=val1 / val0,
                    center_tick=center_tick,
                    upper=upper,
                    lower=lower,
                    upper_delta=upper - center_tick,
                    lower_delta=lower - center_tick,
                )
        idx += tick_spacing
    return None



import os
from datetime import date, timedelta, datetime
from decimal import Decimal
from orjson import orjson
from typing import Tuple, Dict

import numpy as np
import pandas as pd

from ._typing import (
    ETH_MAINNET,
    oSQTH,
    WETH,
    Vault,
    SqueethChain,
    SqueethBalance,
    USDC,
    VaultKey,
    AddVaultAction,
    UpdateCollateralAction,
    UpdateShortAction,
    DepositLpAction,
    WithdrawLpAction,
    ReduceDebtAction,
    LiquidationAction,
    SqueethDescription,
)
from .helper import calc_twap_price, vault_to_dataframe
from .. import MarketInfo, TokenInfo, DemeterError, MarketStatus, DECIMAL_0, UnitDecimal, MarketTypeEnum
from ..broker import Market
from ..data import CacheManager
from ..uniswap import UniLpMarket, PositionInfo
from ..utils import (
    to_decimal,
    float_param_formatter,
    get_formatted_predefined,
    STYLE,
    get_formatted_from_dict,
    console_text,
)


class SqueethMarket(Market):
    """
    | Simulation of squeeth market. Support long and short.
    | In long, you can buy or sell squeeth.
    | In short, you can deposit eth or uniswap lp token to mint squeeth, if collateral is low, vault will be liquidated.

    :param market_info: key of this market
    :type market_info: MarketInfo
    :param squeeth_uni_pool: Instance of oSQTH-ETH uniswap pool, Squeeth depends on this pool to swap.
    :type squeeth_uni_pool: UniLpMarket
    :param data: pool data for back test. downloaded by demeter-fetch
    :type data: pd.DataFrame
    :param data_path: path to load pool data
    :type data_path: str
    """

    def __init__(
        self,
        market_info: MarketInfo,
        squeeth_uni_pool: UniLpMarket,
        data: pd.DataFrame = None,
        data_path: str = "./data",
    ):
        super().__init__(market_info=market_info, data_path=data_path, data=data)
        self._network = ETH_MAINNET
        self._squeeth_uni_pool = squeeth_uni_pool
        self.vault: Dict[VaultKey, Vault] = {}
        self._max_vault_id = 0

    TWAP_PERIOD = 7  # minutes, which is 420 seconds;
    MIN_DEPOSIT_AMOUNT = Decimal("0.5")  # eth
    # the collateralization ratio (CR) is checked with the numerator and denominator separately
    # a user is safe if - collateral value >= (COLLAT_RATIO_NUMER/COLLAT_RATIO_DENOM)* debt value
    CR_NUMERATOR = Decimal(3)
    CR_DENOMINATOR = Decimal(2)
    REDUCE_DEBT_BOUNTY = Decimal("0.02")
    LIQUIDATION_BOUNTY = Decimal("0.1")
    INDEX_SCALE = Decimal(1e4)

    def __str__(self):
        from demeter.utils import orjson_default

        return orjson.dumps(self.description, default=orjson_default).decode()

    @property
    def description(self):
        """
        Get a brief description of this market
        """
        return SqueethDescription(type(self).__name__, self._market_info.name)

    @property
    def osqth_balance(self) -> Decimal:
        """
        Get balance of osqth in your broker account
        :return: osqth balance
        :rtype: Decimal
        """
        return self.broker.get_token_balance(oSQTH)

    @property
    def squeeth_uni_pool(self) -> UniLpMarket:
        """
        Instance of oSQTH-ETH uniswap pool
        :return: oSQTH-ETH uniswap pool
        :rtype: UniLpMarket
        """
        return self._squeeth_uni_pool

    @property
    def network(self) -> SqueethChain:
        """
        Get chain config of this market. Currently, squeeth is only available on Ethereum
        :return: chain config
        :rtype: SqueethChain
        """
        return self._network

    def get_collat_ratio_and_liq_price(self, vault_key: VaultKey) -> Tuple[Decimal, Decimal]:
        """
        | Get collateral ratio and liquidation price of a vault.
        | forked from useGetCollatRatioAndLiqPrice function in hooks.ts

        :param vault_key: key of vault
        :type vault_key: VaultKey
        :return: collateral ratio and liquidation price
        :rtype: Tuple[Decimal, Decimal]
        """
        nf = self.get_norm_factor()
        collateral_in_eth = self._get_effective_collateral_in_eth(vault_key)
        eth_price = self.get_twap_price(WETH)

        debt_amount_in_index = self.vault[vault_key].osqth_short_amount * nf / SqueethMarket.INDEX_SCALE * eth_price
        if debt_amount_in_index == 0:
            return DECIMAL_0, DECIMAL_0
        r_squeeth = self.vault[vault_key].osqth_short_amount * nf / SqueethMarket.INDEX_SCALE
        return collateral_in_eth / debt_amount_in_index, collateral_in_eth / (r_squeeth * Decimal("1.5"))

    def get_market_balance(self) -> SqueethBalance:
        """
        Get current status, including positions, balances

        :param price: external price, is set to none, will use price in current market status
        :type price: pd.Series | Dict[str, Decimal]
        :return: MarketBalance
        :rtype: SqueethBalance
        """
        current_data = self._market_status.data
        price = {oSQTH.name: current_data[oSQTH.name] * current_data[WETH.name], WETH.name: current_data[WETH.name]}

        long_amount = self.osqth_balance
        short_amount = Decimal(sum([x.osqth_short_amount for x in self.vault.values()]))
        short_value = short_amount * price[oSQTH.name]
        eth_price = self.get_twap_price(WETH)
        collateral_eth = Decimal(
            sum(self._get_effective_collateral_in_eth(VaultKey(v.id)) for v in self.vault.values())
        )
        collateral_value = collateral_eth * price[WETH.name]
        short_in_eth = short_amount * self.get_norm_factor() * eth_price / SqueethMarket.INDEX_SCALE
        return SqueethBalance(
            # squeeth token is hold by broker. so net value is only calculated from short position
            net_value=collateral_value - short_value,
            collateral_amount=collateral_eth,
            osqth_long_amount=long_amount,
            osqth_short_amount=short_amount,
            osqth_net_amount=long_amount - short_amount,
            vault_count=len(self.vault),
            delta=Decimal(2) * price[WETH.name],
            gamma=Decimal(2),
            collateral_value=collateral_value,
            osqth_short_in_eth=short_in_eth,
            collateral_ratio=collateral_eth / short_in_eth if short_in_eth != DECIMAL_0 else 0,
        )

    def get_denormalized_mark(self) -> Decimal:
        """
        Get denormalized mark price,
        :return:
        """
        eth_price = self.get_twap_price(WETH)
        osqth_price = self.get_twap_price(oSQTH)
        return eth_price * osqth_price / self.get_norm_factor() * Decimal(1e14)

    def get_index(self) -> Decimal:
        """
        Get index price
        :return:
        """
        self.get_twap_price(WETH)
        return self.get_twap_price(WETH) ** 2 * Decimal(1e10)

    def load_data(self, start_date: date, end_date: date):
        """
        Load data from .minute.csv, then update index and fill null data.

        :param start_date: start test date
        :type start_date: date
        :param end_date: end test date
        :type end_date: date
        """
        cache_key = CacheManager.get_cache_key(self.market_info.type.name, start_date, end_date)
        cache_df = CacheManager.load(cache_key)
        if cache_df is not None:
            self.data = cache_df
            return
        self.logger.info(f"start load files from {start_date} to {end_date}...")
        df = pd.DataFrame()
        day = start_date
        if start_date > end_date:
            raise DemeterError(f"start date {start_date} should earlier than end date {end_date}")
        while day <= end_date:
            path = os.path.join(
                self.data_path,
                f"{self._network.chain.name.lower()}-squeeth-controller-{day.strftime('%Y-%m-%d')}.minute.csv",
            )
            day_df = pd.read_csv(
                path,
                converters={"norm_factor": to_decimal, "WETH": to_decimal, "OSQTH": to_decimal},
            )
            df = pd.concat([df, day_df])
            day = day + timedelta(days=1)
        self.logger.info("load file complete, preparing...")

        df["block_timestamp"] = pd.to_datetime(df["block_timestamp"])
        df.set_index("block_timestamp", inplace=True)
        df = df.ffill()
        if pd.isnull(df.index[0]):
            raise DemeterError(
                f"start date {start_date} does not have available data, Consider start from previous day"
            )
        self.data = df
        CacheManager.save(cache_key, df)
        self.logger.info("data has been prepared")

    def set_market_status(self, market_status: MarketStatus, price: pd.Series | None):
        """
        Set current status (normalize factor, price in uniswap pool) to Market

        :param market_status: market data
        :type market_status: MarketStatus
        :param price: price of token at this moment
        :type price: pd.Series
        """

        super().set_market_status(market_status, price)
        if market_status.data is None:
            market_status.data = self.data.loc[market_status.timestamp]
        self._market_status = market_status

    def get_price_from_data(self) -> pd.DataFrame:
        """
        Extract token price from relative uniswap pool. All price is quoted in usd
        """
        if self.data is None:
            raise DemeterError("data has not set")
        price_df = self._data[[WETH.name, oSQTH.name]].copy()
        price_df[oSQTH.name] = price_df[oSQTH.name] * price_df[WETH.name]
        return price_df

    def formatted_str(self):
        """
        Return a brief description of this market in pretty format. Used for print in console.
        """
        value = get_formatted_predefined(f"{self.market_info.name}({type(self).__name__})", STYLE["header3"]) + "\n"
        balance: SqueethBalance = self.get_market_balance()

        value += (
            get_formatted_from_dict(
                {
                    "net_value": console_text.format_value(balance.net_value),
                    "collateral_amount": console_text.format_value(balance.collateral_amount),
                    "collateral_value": console_text.format_value(balance.collateral_value),
                    "osqth_long_amount": console_text.format_value(balance.osqth_long_amount),
                    "osqth_short_amount": console_text.format_value(balance.osqth_short_amount),
                    "osqth_short_in_eth": console_text.format_value(balance.osqth_short_in_eth),
                    "osqth_net_amount": console_text.format_value(balance.osqth_net_amount),
                    "collateral_ratio": console_text.format_value(balance.collateral_ratio),
                }
            )
            + "\n"
        )
        value += get_formatted_predefined("Vaults", STYLE["key"]) + "\n"
        vault_df = vault_to_dataframe(self.vault)
        value += vault_df.to_string() + "\n" if len(vault_df.index) > 0 else "Empty DataFrame\n"

        return value

    # region short

    @float_param_formatter
    def collateral_amount_to_osqth(
        self, collateral_amount: Decimal | float, collateral_rate: Decimal | float = CR_DENOMINATOR
    ) -> Decimal:
        """
        Calculate osqth amount based on eth amount. Forked from useGetShortAmountFromDebt in hook.ts

        :param collateral_amount: eth amount to collateral
        :type collateral_amount: Decimal | float
        :param collateral_rate: collateral rate, default is 2
        :type collateral_rate: Decimal | float
        :return: osqth amount
        :rtype: Decimal
        """
        eth_price = self.get_twap_price(WETH)
        norm_factor = self.get_norm_factor()
        debt_amount = collateral_amount / collateral_rate
        collateral_in_osqth = debt_amount * SqueethMarket.INDEX_SCALE / norm_factor / eth_price
        return collateral_in_osqth

    @float_param_formatter
    def open_deposit_mint_by_collat_rate(
        self,
        deposit_eth_amount: Decimal | float,
        collateral_rate: Decimal | float = CR_DENOMINATOR,
        vault_key: VaultKey | None = None,
        uni_position: PositionInfo | None = None,
    ) -> Tuple[VaultKey, Decimal]:
        """
        open deposit and mint osqth, you can decide mint amount by collateral_rate,

        :param deposit_eth_amount: eth amount to collate
        :type deposit_eth_amount: Decimal
        :param collateral_rate: collateral_rate
        :type collateral_rate: Decimal
        :param vault_key: vault key, if set to None, will create a new vault. or else append collateral to existing vault
        :type vault_key: VaultKey
        :param uni_position: Uniswap liquid position of osqth-weth pool. You can deposit this position to raise collateral rate
        :type uni_position: PositionInfo
        :return: vault created, and osqth mint amount
        :rtype: Tuple[VaultKey, Decimal]:
        """
        osqth_amount = self.collateral_amount_to_osqth(deposit_eth_amount, collateral_rate)
        return self.open_deposit_mint(deposit_eth_amount, osqth_amount, vault_key, uni_position)

    @float_param_formatter
    def open_deposit_mint(
        self,
        deposit_eth_amount: Decimal | float,
        osqth_mint_amount: Decimal | float = DECIMAL_0,
        vault_key: VaultKey | None = None,
        uni_position: PositionInfo | None = None,
    ) -> Tuple[VaultKey, Decimal]:
        """
        open deposit and mint osqth, Forked function _openDepositMint in controller.sol,

        :param deposit_eth_amount: eth amount to collate
        :type deposit_eth_amount: Decimal
        :param osqth_mint_amount: osqth mint amount
        :type osqth_mint_amount: Decimal
        :param vault_key: vault key, if set to None, will create a new vault. or else append collateral to existing vault
        :type vault_key: VaultKey
        :param uni_position: Uniswap liquid position of osqth-weth pool. You can deposit this position to raise collateral rate
        :type uni_position: PositionInfo
        :return: vault created, and osqth mint amount
        :rtype: Tuple[VaultKey, Decimal]:
        """

        norm_factor = self.get_norm_factor()
        deposit_amount_with_fee = deposit_eth_amount
        if vault_key is None:
            self._max_vault_id += 1
            vault_key = VaultKey(self._max_vault_id)
            self.vault[vault_key] = Vault(vault_key.id)
            self._record_action(
                AddVaultAction(
                    market=self.market_info,
                    vault_id=vault_key.id,
                    vault_count=len(self.vault),
                )
            )
        fee_amount = Decimal(0)
        if osqth_mint_amount > DECIMAL_0:
            fee_amount, deposit_amount_with_fee = self._get_fee(
                self.vault[vault_key], deposit_eth_amount, osqth_mint_amount
            )
            self.vault[vault_key].osqth_short_amount += osqth_mint_amount
            self.broker.add_to_balance(oSQTH, osqth_mint_amount)
            self._record_action(
                UpdateShortAction(
                    market=self._market_info,
                    vault_id=vault_key.id,
                    short_amount=UnitDecimal(osqth_mint_amount, oSQTH.name),
                    short_after=UnitDecimal(self.vault[vault_key].osqth_short_amount, oSQTH.name),
                )
            )

        if deposit_eth_amount > 0:
            self.deposit(vault_key, deposit_amount_with_fee)

        if uni_position is not None:
            self._deposit_uni_position(vault_key, uni_position)

        # check vault
        self._check_vault(vault_key, norm_factor)

        if fee_amount > 0:
            self.broker.subtract_from_balance(WETH, fee_amount)

        return vault_key, osqth_mint_amount

    def _check_vault(self, vault_key: VaultKey, norm_factor: Decimal):
        """
        Check vault is safe (collateral is above 150%), or is above minimal collateral amount(0.5eth)

        :param vault_key: Which vault to operate.
        :type vault_key: VaultKey
        :param norm_factor: normalize factor
        :type norm_factor: Decimal
        :return:
        """
        is_safe, is_dust = self.get_vault_status(vault_key, norm_factor)
        if not is_safe:
            raise DemeterError("Vault collateral rate is not safe")
        if is_dust:
            raise DemeterError("Vault collateral is below dust")

    def get_vault_status(
        self, vault_key: VaultKey, norm_factor: Decimal, twap_eth_price: Decimal | None = None
    ) -> Tuple[bool, bool]:
        """
        Check vault is safe (collateral is above 150%), or is above minimal collateral amount(0.5eth)

        :param vault_key: Which vault to operate.
        :type vault_key: VaultKey
        :param norm_factor: normalize factor
        :type norm_factor: Decimal
        :param twap_eth_price: twap eth price.
        :type twap_eth_price: Decimal
        :return: vault status, is safe, is below min-amount
        :rtype: Tuple[bool, bool]
        """
        if twap_eth_price is None:
            twap_eth_price = self.get_twap_price(WETH)
        if self.vault[vault_key].osqth_short_amount == 0:
            return True, False

        debt_value_in_eth = (
            self.vault[vault_key].osqth_short_amount * norm_factor * twap_eth_price / SqueethMarket.INDEX_SCALE
        )
        total_collateral = self._get_effective_collateral_in_eth(vault_key, norm_factor, twap_eth_price)

        is_dust = total_collateral < SqueethMarket.MIN_DEPOSIT_AMOUNT
        is_above_water = (
            total_collateral * SqueethMarket.CR_DENOMINATOR >= debt_value_in_eth * SqueethMarket.CR_NUMERATOR
        )

        return is_above_water, is_dust

    def _get_effective_collateral_in_eth(
        self, vault_key: VaultKey, norm_factor: Decimal | None = None, eth_price: Decimal | None = None
    ) -> Decimal:
        """
        | Get total collateral amount, including: eth deposited, uniswap lp value(eth and osqth)
        | Note: Osqth amount is converted to eth by index price.
        """
        if self.vault[vault_key].uni_nft_id is None:
            return self.vault[vault_key].collateral_amount
        if norm_factor is None:
            norm_factor = self.get_norm_factor()
        if eth_price is None:
            eth_price = self.get_twap_price(WETH)
        position_info = self.vault[vault_key].uni_nft_id
        nft_weth_amount, nft_squeeth_amount = self.squeeth_uni_pool.get_position_amount(position_info)
        fee_weth = self.squeeth_uni_pool.positions[position_info].pending_amount0
        fee_squeeth = self.squeeth_uni_pool.positions[position_info].pending_amount1
        nft_weth_amount += fee_weth
        nft_squeeth_amount += fee_squeeth
        # IMPORTANT:
        # According to _getEffectiveCollateral function in controller.sol,
        # we calculate osqth amount to eth amount by index price,
        # so it's different to net value calculated by uniswap pool(who uses mark price).
        osqth_index_val_in_eth = nft_squeeth_amount / SqueethMarket.INDEX_SCALE * norm_factor * eth_price

        return nft_weth_amount + osqth_index_val_in_eth + self.vault[vault_key].collateral_amount

    def get_twap_price(self, token: TokenInfo, now: datetime | None = None) -> Decimal:
        """
        | Get twap(time weighted average price) price, Just like what uniswap oracle contract did.
        | Depends on price stored in self.data

        :param token: Which token to calculate
        :type token: TokenInfo
        :param now: end time.
        :type now: datetime
        :return: twap price
        :rtype: Decimal
        """
        # for test case
        if self._market_status.timestamp is None:
            return self._market_status.data[token.name]
        if now is None:
            now = self._market_status.timestamp
        start = now - timedelta(minutes=SqueethMarket.TWAP_PERIOD - 1)
        if start < self.data.index[0]:
            start = self.data.index[0].to_pydatetime()
        # remember 1 minute has 1 data point
        prices: pd.Series = self.data[start:now][token.name]
        return calc_twap_price(prices)

    def _get_fee(
        self, vault: Vault, deposit_eth_amount: Decimal, osqth_mint_amount: Decimal
    ) -> Tuple[Decimal, Decimal]:
        """
        Get fee of deposit, currently fee rate is 0 and unlikely to change, so this part is omitted.
        """
        return Decimal(0), deposit_eth_amount

    @float_param_formatter
    def deposit(self, vault_key: VaultKey, eth_value: Decimal | float):
        """
        Deposit eth to current vault

        :param vault_key: Vault key
        :type vault_key: VaultKey
        :param eth_value: Eth amount to collate
        :type eth_value: Decimal
        """
        self.vault[vault_key].collateral_amount += eth_value
        self.broker.subtract_from_balance(WETH, eth_value)
        self._record_action(
            UpdateCollateralAction(
                market=self.market_info,
                vault_id=vault_key.id,
                collateral_amount=UnitDecimal(eth_value, WETH.name),
                collateral_after=UnitDecimal(self.vault[vault_key].collateral_amount, WETH.name),
                fee=UnitDecimal(DECIMAL_0, WETH.name),
            )
        )

    def deposit_uni_position(self, vault_key: VaultKey, uni_position_info: PositionInfo):
        """
        Deposit uniswap lp position to vault. After deposit, position will be transferred to this market.
        and its net value will not be counted in uniswap market.

        :param vault_key: Vault key
        :type vault_key: VaultKey
        :param uni_position_info: Uniswap lp position
        :type uni_position_info: PositionInfo
        """
        self._deposit_uni_position(vault_key, uni_position_info)

    def _deposit_uni_position(self, vault_key: VaultKey, uni_position: PositionInfo):
        if uni_position not in self.squeeth_uni_pool.positions:
            raise DemeterError("Position is not in squeeth-eth pool")
        if self.squeeth_uni_pool.positions[uni_position].liquidity <= 0:
            raise DemeterError("Require liquidity in squeeth-eth pool")
        if self.vault[vault_key].uni_nft_id is not None:
            raise DemeterError("This vault already has a NFT collateral")
        self.vault[vault_key].uni_nft_id = uni_position
        # transfer lp position to current market.
        self.squeeth_uni_pool.transfer_position_out(uni_position)
        self._record_action(
            DepositLpAction(
                market=self.market_info,
                vault_id=vault_key.id,
                position=uni_position,
            )
        )

    def _withdraw_collateral(self, vault_key: VaultKey, amount: Decimal | None = None):
        if vault_key not in self.vault:
            raise DemeterError(f"{vault_key.id} not exist")
        if amount is None or amount > self.vault[vault_key].collateral_amount:
            amount = self.vault[vault_key].collateral_amount

        self.vault[vault_key].collateral_amount -= amount
        self.broker.add_to_balance(WETH, amount)
        self._check_vault(vault_key, self.get_norm_factor())
        self._record_action(
            UpdateCollateralAction(
                market=self.market_info,
                vault_id=vault_key.id,
                collateral_amount=UnitDecimal(DECIMAL_0 - amount, WETH.name),
                collateral_after=UnitDecimal(self.vault[vault_key].collateral_amount, WETH.name),
                fee=UnitDecimal(DECIMAL_0, WETH.name),
            )
        )

    def withdraw_uni_position(self, vault_key: VaultKey, uni_position: PositionInfo):
        """
        Withdraw uniswap lp position.

        :param vault_key: Vault key
        :type vault_key: VaultKey
        :param uni_position_info: Uniswap lp position
        :type uni_position_info: PositionInfo
        """
        if vault_key not in self.vault:
            raise DemeterError(f"{vault_key.id} not exist")
        if self.vault[vault_key].uni_nft_id != uni_position:
            raise DemeterError(f"{uni_position} is not deposit in vault {vault_key.id}")
        self.vault[vault_key].uni_nft_id = None
        self.squeeth_uni_pool.transfer_position_in(uni_position)
        self._check_vault(vault_key, self.get_norm_factor())
        self._record_action(
            WithdrawLpAction(
                market=self.market_info,
                vault_id=vault_key.id,
                position=uni_position,
            )
        )

    @float_param_formatter
    def burn_and_withdraw(
        self, vault_key: VaultKey, osqth_burn_amount: Decimal | float, withdraw_eth_amount: Decimal | float
    ):
        """
        Burn osqth and withdraw. uniswap lp position is not handled here.

        :param vault_key: Vault key
        :type vault_key: VaultKey
        :param osqth_burn_amount: osqth to burn
        :type osqth_burn_amount: Decimal
        :param withdraw_eth_amount: eth to withdraw
        :type withdraw_eth_amount: Decimal
        :return:
        """
        if vault_key not in self.vault:
            raise DemeterError(f"{vault_key.id} not exist")
        vault = self.vault[vault_key]
        if osqth_burn_amount > 0:
            if vault.osqth_short_amount >= osqth_burn_amount:
                removed_amount = osqth_burn_amount
                vault.osqth_short_amount -= osqth_burn_amount
            else:
                removed_amount = vault.osqth_short_amount
                vault.osqth_short_amount = 0
            self.broker.subtract_from_balance(oSQTH, removed_amount)
            self._record_action(
                UpdateShortAction(
                    market=self.market_info,
                    vault_id=vault_key.id,
                    short_amount=UnitDecimal(DECIMAL_0 - removed_amount, oSQTH.name),
                    short_after=UnitDecimal(vault.osqth_short_amount, oSQTH.name),
                )
            )

        if withdraw_eth_amount > 0:
            self._withdraw_collateral(vault_key, withdraw_eth_amount)

        self._check_vault(vault_key, self.get_norm_factor())

    # endregion

    # region long
    @float_param_formatter
    def buy_squeeth(
        self, osqth_amount: float | Decimal | None = None, eth_amount: float | Decimal | None = None
    ) -> Tuple[Decimal, Decimal, Decimal]:
        """
        Buy osqth, you can specify osqth amount to buy. or how many eth to swap to osqth

        :param osqth_amount: osqth amount to buy
        :type osqth_amount: Decimal
        :param eth_amount: eth amount to swap to osqth
        :type eth_amount: Decimal
        :return: swap fee, eth spent, osqth got
        :rtype: Tuple[Decimal, Decimal, Decimal]
        """
        if osqth_amount is None and eth_amount is not None:
            osqth_amount = eth_amount / self._market_status.data["OSQTH"]
        fee, eth_amount, osqth_amount = self._squeeth_uni_pool.buy(osqth_amount)

        return fee, eth_amount, osqth_amount

    @float_param_formatter
    def sell_squeeth(
        self, osqth_amount: float | Decimal | None = None, eth_amount: float | Decimal | None = None
    ) -> Tuple[Decimal, Decimal, Decimal]:
        """
        Sell osqth, you can specify osqth amount to sell. or how many eth to get

        :param osqth_amount: osqth amount to sell
        :type osqth_amount: Decimal
        :param eth_amount: eth amount to got
        :type eth_amount: Decimal
        :return: swap fee,  osqth sold,eth got,
        :rtype: Tuple[Decimal, Decimal, Decimal]
        """
        if osqth_amount is None and eth_amount is not None:
            osqth_amount = eth_amount / self._market_status.data["OSQTH"]
        fee, osqth_amount, eth_amount = self._squeeth_uni_pool.sell(osqth_amount)
        return fee, osqth_amount, eth_amount

    # endregion

    # region liquidate

    def update(self):
        """
        Check and trigger liquidation
        """
        norm_factor = self.get_norm_factor()
        eth_price = self.get_twap_price(WETH)
        for vk, v in self.vault.items():
            is_above_water, is_dust = self.get_vault_status(vk, norm_factor, eth_price)
            if not is_above_water:
                self.liquidate(vk)

    def liquidate(self, vault_key: VaultKey) -> Decimal:
        """
        | liquidate if collateral ratio is blow 150%.
        | liquidate process will first try to redeem unswap lp position, then burn collected osqth, and deposit collected eth.
        | if the vault is still not safe, will liquidate 50% osqth debt. you will have to pay corresponding collateral, and liquidate bounty.
        | If after liquidation, collateral is blow minial amount(0.5eth), will liquidate all debt.

        :param vault_key: vault key
        :type vault_key: VaultKey
        :return: Liquidated debt amount
        :rtype: Decimal
        """
        if vault_key not in self.vault:
            raise DemeterError(f"{vault_key.id} not exist")
        norm_factor = self.get_norm_factor()
        vault = self.vault[vault_key]
        is_safe, is_dust = self.get_vault_status(vault_key, norm_factor)
        if is_safe:
            raise DemeterError("Can not liquidate safe vault")
        # try to save target vault before liquidation by reducing debt
        osqth_burn_amount, osqth_excess, bounty, eth_collected = self._reduce_debt(vault_key, True)

        is_safe, is_dust = self.get_vault_status(vault_key, self.get_norm_factor())
        if is_safe:
            # should transfer bounty to liquidater
            return DECIMAL_0

        vault.collateral_amount += bounty  # add bounty back, will re-calculate liq bounty
        debt_amount, collateral_paid = self._liquidate(vault, vault.osqth_short_amount, self.get_norm_factor())
        return debt_amount

    def _liquidate(self, vault: Vault, max_debt_amount: Decimal, norm_factor: Decimal) -> Tuple[Decimal, Decimal]:
        liquidate_amount, collateral_to_pay = self._get_liquidation_result(
            max_debt_amount, vault.osqth_short_amount, vault.collateral_amount
        )
        # if the liquidator didn't specify enough wPowerPerp to burn, revert.
        if max_debt_amount < liquidate_amount:
            raise DemeterError("Need full liquidation")

        vault.osqth_short_amount -= liquidate_amount
        vault.collateral_amount -= collateral_to_pay

        is_safe, is_dust = self.get_vault_status(VaultKey(vault.id), norm_factor)
        if is_dust:
            raise DemeterError("Dust vault left")
        self._record_action(
            LiquidationAction(
                market=self.market_info,
                vault_id=vault.id,
                liquidate_amount=UnitDecimal(liquidate_amount, oSQTH.name),
                short_amount_after=UnitDecimal(vault.osqth_short_amount, oSQTH.name),
                collateral_to_pay=UnitDecimal(collateral_to_pay, WETH.name),
                collateral_after=UnitDecimal(vault.collateral_amount, WETH.name),
            )
        )
        return liquidate_amount, collateral_to_pay

    def _get_liquidation_result(
        self, max_osqth_amount: Decimal, vault_short_amount: Decimal, vault_collateral_amount: Decimal
    ) -> Tuple[Decimal, Decimal]:
        final_liquidate_amount, collateral_to_pay = self._get_single_liquidation_amount(
            max_osqth_amount, vault_short_amount / 2
        )

        if vault_collateral_amount > collateral_to_pay:
            if vault_collateral_amount - collateral_to_pay < SqueethMarket.MIN_DEPOSIT_AMOUNT:
                # the vault is left with dust after liquidation, allow liquidating full vault
                # calculate the new liquidation amount and collateral again based on the new limit
                final_liquidate_amount, collateral_to_pay = self._get_single_liquidation_amount(
                    max_osqth_amount, vault_short_amount
                )

        # check if final collateral to pay is greater than vault amount.
        # if so the system only pays out the amount the vault has, which may not be profitable
        if collateral_to_pay > vault_collateral_amount:
            final_liquidate_amount = vault_short_amount
            collateral_to_pay = vault_collateral_amount

        return final_liquidate_amount, collateral_to_pay

    def _get_single_liquidation_amount(
        self, max_input_osqth: Decimal, max_liquidatable_osqth: Decimal
    ) -> Tuple[Decimal, Decimal]:
        final_amount = max_liquidatable_osqth if max_input_osqth > max_liquidatable_osqth else max_input_osqth

        osqth_price = self.get_twap_price(oSQTH)
        collateral_to_pay: Decimal = final_amount * osqth_price

        # add 10% bonus for liquidators
        collateral_to_pay += collateral_to_pay * SqueethMarket.LIQUIDATION_BOUNTY

        return final_amount, collateral_to_pay

    def _reduce_debt(self, vault_key: VaultKey, pay_bounty: bool) -> Tuple[Decimal, Decimal, Decimal, Decimal]:
        """
        Redeem uni lp, and burn those osqth, to save the account
        """
        vault = self.vault[vault_key]

        if vault.uni_nft_id is None:
            return DECIMAL_0, DECIMAL_0, DECIMAL_0, DECIMAL_0
        position = vault.uni_nft_id
        withdrawn_eth_amount, withdrawn_osqth_amount = self._redeem_uni_token(position)
        osqth_burn_amount, osqth_excess, bounty = self._get_reduce_debt_result_in_vault(
            vault, withdrawn_eth_amount, withdrawn_osqth_amount, pay_bounty
        )
        if osqth_excess > 0:
            self.broker.add_to_balance(oSQTH, osqth_excess)

        self._record_action(
            ReduceDebtAction(
                market=self.market_info,
                vault_id=vault_key.id,
                position=position,
                withdrawn_eth_amount=UnitDecimal(withdrawn_eth_amount, WETH.name),
                withdrawn_osqth_amount=UnitDecimal(withdrawn_osqth_amount, oSQTH.name),
                burn_amount=UnitDecimal(osqth_burn_amount, oSQTH.name),
                excess=UnitDecimal(osqth_excess, oSQTH.name),
                bounty=UnitDecimal(bounty, WETH.name),
                short_amount_after=UnitDecimal(vault.osqth_short_amount, oSQTH.name),
                collateral_after=UnitDecimal(vault.collateral_amount, WETH.name),
            )
        )
        return osqth_burn_amount, osqth_excess, bounty, withdrawn_eth_amount

    def _get_reduce_debt_result_in_vault(
        self, vault: Vault, nft_eth_amount: Decimal, nft_osqth_amount: Decimal, pay_bounty: bool
    ):
        bounty = DECIMAL_0
        if pay_bounty:
            bounty = self._get_reduce_debt_bounty(nft_eth_amount, nft_osqth_amount)

        burn_amount = nft_osqth_amount
        osqth_excess = DECIMAL_0
        if nft_osqth_amount > vault.osqth_short_amount:
            osqth_excess = nft_osqth_amount - vault.osqth_short_amount
            burn_amount = vault.osqth_short_amount

        vault.osqth_short_amount -= burn_amount
        vault.uni_nft_id = None
        vault.collateral_amount += nft_eth_amount
        vault.collateral_amount -= bounty

        return burn_amount, osqth_excess, bounty

    def _get_reduce_debt_bounty(self, eth_withdrawn: Decimal, osqth_reduced: Decimal) -> Decimal:
        price = self.get_twap_price(oSQTH)
        return (osqth_reduced * price + eth_withdrawn) * Decimal(SqueethMarket.REDUCE_DEBT_BOUNTY)

    def _redeem_uni_token(self, position_info: PositionInfo) -> Tuple[Decimal, Decimal]:
        self.squeeth_uni_pool.remove_liquidity(position_info, collect=False)
        weth_get, osqth_get = self.squeeth_uni_pool.collect_fee(position_info, collect_to_user=False)
        return weth_get, osqth_get

    # endregion
    def get_norm_factor(self) -> Decimal:
        """
        Get current normalize factor. It is downloaded from event log in squeeth contract.

        :return: normalize factor at current time
        :rtype: Decimal
        """
        # Maybe I should calculate this myself, as transactions are too few in a day
        return self._market_status.data["norm_factor"]

    def _resample(self, freq: str):
        self._data = self.data.resample(freq).first()



"""
Uniswap market, this module can simulate common operations in Uniswap V3, such as add/remove liquidity, swap etc.
"""

from ._typing import (
    UniV3Pool,
    UniV3PoolStatus,
    Position,
    UniLpBalance,
    PositionInfo,
    SellAction,
    BuyAction,
    RemoveLiquidityAction,
    CollectFeeAction,
    AddLiquidityAction,
    UniswapMarketStatus,
    UniDescription,
)
from .core import V3CoreLib
from .data import LineTypeEnum, UniLPData
from .market import UniLpMarket
from .helper import (
    nearest_usable_tick,
    tick_to_base_unit_price,
    tick_to_sqrt_price_x96,
    sqrt_price_x96_to_tick,
    sqrt_price_x96_to_base_unit_price,
    get_sqrt_ratio_at_tick,
    from_atomic_unit,
    base_unit_price_to_tick,
    base_unit_price_to_sqrt_price_x96,
    get_swap_value,
    get_swap_value_with_part_balance_used,
)



import pandas as pd
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import NamedTuple, List, Union, Tuple

from demeter import MarketStatus, BaseAction
from demeter._typing import MarketDescription, DemeterError
from demeter.broker import MarketBalance, ActionTypeEnum
from demeter.utils.console_text import get_action_str, ForColorEnum


@dataclass
class DeribitOptionDescription(MarketDescription):
    """
    Designed to generate json description for aave market

    :param type: market type
    :type type: str
    :param name: market name
    :type name: str
    :param position_count: count of positions
    :type position_count: int

    """

    token: str
    positions: List[str]
    """count of position"""


@dataclass
class InstrumentStatus:
    """
    Property of an instrument(order)

    :param: state: status of an instrument(open, closed)
    :type state: str | None
    :param: type: order type, CALL/PUT
    :type type: str | None
    :param: strike_price: strike price
    :type strike_price: int | None
    :param: t: Time remaining until expiration
    :type t: float | None
    :param: expiry_time: expiration time
    :type expiry_time: datetime | None
    :param: vega: greeks, vega
    :type vega: float | None
    :param: theta: greeks, theta
    :type theta: float | None
    :param: rho: greeks, rho
    :type rho: float | None
    :param: gamma: greeks, gamma
    :type gamma: float | None
    :param: delta: greeks, delta
    :type delta: float | None
    :param: underlying_price: underlying price for implied volatility calculations, if token is eth, the unit is eth
    :type underlying_price: float | None
    :param: settlement_price: The settlement price for the instrument, if token is eth, the unit is eth
    :type settlement_price: float | None
    :param: mark_price: The mark price for the instrument, if token is eth, the unit is eth
    :type mark_price: float | None
    :param: mark_iv: implied volatility for mark price
    :type mark_iv: float | None
    :param: last_price: The price for the last trade, if token is eth, the unit is eth
    :type last_price: float | None
    :param: interest_rate: Interest rate used in implied volatility calculations
    :type interest_rate:  float | None
    :param: bid_iv:  implied volatility for best bid
    :type bid_iv: float | None
    :param: best_bid_price: The current best bid price, null if there aren't any bids, if token is eth, the unit is eth
    :type best_bid_price: float | None
    :param: best_bid_amount: It represents the requested order size of all best bids
    :type best_bid_amount: float | None
    :param: ask_iv: implied volatility for best ask
    :type ask_iv: float | None
    :param: best_ask_price: The current best ask price, null if there aren't any asks, if token is eth, the unit is eth
    :type best_ask_price: float | None
    :param: best_ask_amount: It represents the requested order size of all best asks
    :type best_ask_amount: float | None
    :param: asks: List of asks, in price and amount
    :type asks: List[Tuple[float, float]] | None
    :param: bids: List of bids, in price and amount
    :type bids: List[Tuple[float, float]] | None
    """

    state: str | None = None
    type: str | None = None
    strike_price: int | None = None
    t: float | None = None
    expiry_time: datetime | None = None
    vega: float | None = None
    theta: float | None = None
    rho: float | None = None
    gamma: float | None = None
    delta: float | None = None
    underlying_price: float | None = None
    settlement_price: float | None = None
    mark_price: float | None = None
    mark_iv: float | None = None
    last_price: float | None = None
    interest_rate: float | None = None
    bid_iv: float | None = None
    best_bid_price: float | None = None
    best_bid_amount: float | None = None
    ask_iv: float | None = None
    best_ask_price: float | None = None
    best_ask_amount: float | None = None
    asks: List[Tuple[float, float]] | None = None  # price and amount
    bids: List[Tuple[float, float]] | None = None


class DeribitTokenConfig(NamedTuple):
    """
    Config for token in deribit

    :param trade_fee_rate: fee rate for take or make order
    :type trade_fee_rate: Decimal
    :param delivery_fee_rate: fee rate for delivery when exercising
    :type delivery_fee_rate: Decimal
    :param min_trade_decimal: minimal decimal for trade
    :type min_trade_decimal: int
    :param min_fee_decimal: minimal decimal for fee
    :type min_fee_decimal: int
    """

    trade_fee_rate: Decimal
    delivery_fee_rate: Decimal
    min_trade_decimal: int  # exponent, eg: 5 in 1e5, -2 in 1e-2
    min_fee_decimal: int

    @property
    def min_amount(self):
        return Decimal(f"1e{self.min_trade_decimal}")


class OptionKind(Enum):
    """
    Option kind, call/put
    """

    put = "PUT"
    call = "CALL"


@dataclass
class OptionPosition:
    """
    Information for an option position

    :param instrument_name: instrument name
    :type instrument_name: str
    :param expiry_time: expiry time
    :type expiry_time: datetime
    :param strike_price: strike price, if token is eth, the unit is eth
    :type strike_price: int
    :param type: option type
    :type type: OptionKind
    :param amount: amount of a instrument
    :type amount: Decimal
    :param avg_buy_price: average buy price, if token is eth, the unit is eth
    :type avg_buy_price: Decimal
    :param buy_amount: total amount brought
    :type buy_amount: Decimal
    :param avg_sell_price: average sell price, if token is eth, the unit is eth
    :type avg_sell_price: Decimal
    :param sell_amount:  total amount sold
    :type sell_amount: Decimal

    """

    instrument_name: str
    expiry_time: datetime
    strike_price: int
    type: OptionKind
    amount: Decimal
    avg_buy_price: Decimal
    buy_amount: Decimal
    avg_sell_price: Decimal
    sell_amount: Decimal


@dataclass
class DeribitMarketStatus(MarketStatus):
    """
    MarketStatus properties

    :param data: current pool status
    :type data: Union[pd.Series, UniV3PoolStatus]
    """

    data: Union[pd.DataFrame, List[InstrumentStatus]] = None


@dataclass
class OptionMarketBalance(MarketBalance):
    """
    Balance of an option market

    :param puts: List of put option position
    :type puts: List[str]
    :param calls: List of call option position
    :type calls: List[str]
    :param delta: delta of current exposure
    :type delta: Decimal
    :param gamma: gamma of current exposure
    :type gamma: Decimal
    """

    balance: Decimal
    premium: Decimal
    delta: Decimal
    gamma: Decimal


class Order(NamedTuple):
    """
    Order instance when buy/sell option

    :param price: price, if token is eth, the unit is eth
    :type price: Decimal
    :param amount: amount to buy
    :type amount: Decimal
    """

    price: Decimal
    amount: Decimal

    @staticmethod
    def get_average_price(orders: List):
        total = sum([t.amount * Decimal(t.price) for t in orders])
        total_amount = sum([t.amount for t in orders])
        if total_amount == 0:
            return 0
        else:
            return total / total_amount


@dataclass
class OptionTradeAction(BaseAction):
    """
    Recording information about options trades

    :param instrument_name: instrument name
    :type instrument_name: str
    :param type: option type, call/put
    :type type: OptionKind
    :param average_price: average price, if token is eth, the unit is eth
    :type average_price: Decimal
    :param amount: amount to trade
    :type amount: Decimal
    :param total_premium: total premium of this trade
    :type total_premium: Decimal
    :param mark_price: mark price
    :type mark_price: Decimal
    :param underlying_price: underlying price
    :type underlying_price: Decimal
    :param fee: fee of this trade
    :type fee: Decimal
    :param orders: orders in price-amount,
    :type orders: List[Order]
    """

    instrument_name: str
    type: OptionKind
    average_price: Decimal
    amount: Decimal
    total_premium: Decimal
    mark_price: Decimal
    underlying_price: Decimal
    fee: Decimal
    orders: List[Order]

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "instrument_name": self.instrument_name,
                "average_price": str(self.average_price),
                "amount": str(self.amount),
                "total_premium": str(self.total_premium),
                "underlying_price": str(self.underlying_price),
            },
        )


@dataclass
class BuyAction(OptionTradeAction):
    """
    buy action
    """

    def set_type(self):
        self.action_type = ActionTypeEnum.option_buy


@dataclass
class SellAction(OptionTradeAction):
    """
    sell action
    """

    def set_type(self):
        self.action_type = ActionTypeEnum.option_sell


@dataclass
class DepositAction(BaseAction):
    token: str
    amount: Decimal

    def set_type(self):
        self.action_type = ActionTypeEnum.deribit_deposit

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "token": self.token,
                "amount": str(self.amount),
            },
        )


@dataclass
class WithdrawAction(BaseAction):
    token: str
    amount: Decimal

    def set_type(self):
        self.action_type = ActionTypeEnum.deribit_withdraw

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "token": self.token,
                "amount": str(self.amount),
            },
        )


@dataclass
class DeliverAction(BaseAction):
    """
    :param instrument_name: instrument name
    :type instrument_name: str
    :param type: option type, call/put
    :type type: OptionKind
    :param mark_price: mark price, if token is eth, the unit is eth
    :type mark_price: Decimal
    :param amount: total amount
    :type amount: Decimal
    :param total_premium: Decimal
    :type total_premium: Decimal
    :param strike_price: strike price
    :type strike_price: int
    :param underlying_price: underlying price
    :type underlying_price: Decimal
    :param deriver_amount: deriver amount
    :type deriver_amount: Decimal
    :param fee: fee
    :type fee: Decimal
    :param income_amount: income amount
    :type income_amount: Decimal
    """

    instrument_name: str
    type: OptionKind
    mark_price: Decimal
    amount: Decimal
    total_premium: Decimal
    strike_price: int
    underlying_price: Decimal
    deriver_amount: Decimal
    fee: Decimal
    income_amount: Decimal

    def set_type(self):
        self.action_type = ActionTypeEnum.option_deliver

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "instrument_name": str(self.instrument_name),
                "deriver_amount": str(self.deriver_amount),
                "total_premium": str(self.total_premium),
                "underlying_price": str(self.underlying_price),
            },
        )


@dataclass
class ExpiredAction(BaseAction):
    """
    :param instrument_name: instrument name
    :type instrument_name: str
    :param type: option type, call/put
    :type type: OptionKind
    :param mark_price: mark price
    :type mark_price: Decimal
    :param amount: total amount
    :type amount: Decimal
    :param total_premium: Decimal
    :type total_premium: Decimal
    :param strike_price: strike price
    :type strike_price: int
    :param underlying_price: underlying price
    :type underlying_price: Decimal
    """

    instrument_name: str
    type: OptionKind
    mark_price: Decimal
    amount: Decimal
    total_premium: Decimal
    strike_price: int
    underlying_price: Decimal

    def set_type(self):
        self.action_type = ActionTypeEnum.option_expire

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "instrument_name": str(self.instrument_name),
                "underlying_price": str(self.underlying_price),
                "amount": str(self.amount),
                "total_premium": str(self.total_premium),
            },
        )


class InsufficientBalanceError(DemeterError):
    def __init__(self, message):
        self.message = message


DERIBIT_OPTION_FREQ = "1h"



"""
Utils functions of demeter
"""

from .application import (
    float_param_formatter,
    to_decimal,
    to_multi_index_df,
    load_account_status,
    orjson_default,
    require,
)
from .console_text import (
    ForColorEnum,
    BackColorEnum,
    ModeEnum,
    get_formatted,
    STYLE,
    get_formatted_predefined,
    get_formatted_from_dict,
)



from ._typing import MetricEnum
from .core import performance_metrics, round_results
from .calculator import (
    return_value,
    return_rate,
    return_multiple,
    return_rate_series,
    annualized_return,
    max_draw_down,
    volatility,
    sharpe_ratio,
    alpha_beta,
)



import os.path
from demeter import RowData, MarketStatus
from datetime import date, timedelta
from decimal import Decimal, ROUND_DOWN
from typing import List, Set, Dict, Tuple, Union
import pandas as pd
from orjson import orjson
from ..broker import Market, MarketInfo
from .typing import GmxDescription, TokenInfo, GmxBalance, BuyGlpAction, SellGlpAction
from .._typing import ChainType, DemeterError
from ..utils import to_decimal


class GmxMarket(Market):
    """
    GMX Market is the simulator for the GMX, here you can simulate some transactions
    """

    def __init__(self,
                 market_info: MarketInfo,
                 data: pd.DataFrame = None,
                 data_path: str = "./data",
                 tokens: List[TokenInfo] = None,):
        super().__init__(market_info=market_info, data=data, data_path=data_path)
        tokens = tokens if tokens is not None else []
        self.glp_amount = Decimal("0.00")  # glp liquidity
        self.glp_decimal = 18
        self.PRICE_PRECISION = 10 ** 30
        self.reward = Decimal("0.00")  # pending fee
        self.mint_burn_fee_basis_points = 25
        self.tax_basis_points = 60
        self._tokens: Set[TokenInfo] = set()
        self.add_token(tokens)

    def add_token(self, token_info: TokenInfo | List[TokenInfo]):
        """
        Add one or an array of token to aave back test.

        :param token_info: tokens to add
        :type token_info: TokenInfo | List[TokenInfo]
        """
        if not isinstance(token_info, list):
            token_info = [token_info]
        for t in token_info:
            self._tokens.add(t)

    def __str__(self):
        from demeter.utils import orjson_default
        return orjson.dumps(self.description, default=orjson_default).decode()

    def description(self):
        return GmxDescription(
            type=type(self).__name__,
            name=self._market_info.name
        )

    def update(self):
        self._update_fee()

    def _update_fee(self):
        block_reward = Decimal(self.market_status.data.interval) * 60
        supply = Decimal(self.market_status.data.glp)
        reward = block_reward * self.glp_amount / supply
        self.reward += reward

    def buy_glp(self,
                token: TokenInfo,
                amount: Decimal | float):
        """
        token: which token to buy glp
        amount: how much token to buy glp
        """
        glp_amount = self._add_liquidity(token, amount)
        self.glp_amount += glp_amount
        self.broker.subtract_from_balance(token, amount)
        return glp_amount

    def sell_glp(self,
                 token: TokenInfo,
                 glp_amount: Decimal | float = 0):
        """
        sell glp to get token
        """
        if not glp_amount:
            glp_amount = self.glp_amount
        token_amount = self._remove_liquidity(token, glp_amount)
        self.glp_amount -= glp_amount
        self.broker.add_to_balance(token, token_amount)
        return token_amount

    def _add_liquidity(self,
                       token: TokenInfo,
                       amount: Decimal | float):
        glp_supply = Decimal(self.market_status.data.glp)
        aum = Decimal(self.market_status.data.aum)
        aum_in_usdg = aum / Decimal(10 ** 12)
        aum_in_usdg = aum_in_usdg.quantize(Decimal('0'), rounding=ROUND_DOWN)
        usdg_amount = self.buy_usdg(token, amount)  # token amount convert to usdg amount
        mint_amount = usdg_amount * glp_supply / aum_in_usdg
        mint_amount = mint_amount.quantize(Decimal('0'), rounding=ROUND_DOWN)
        self._record_action(
            BuyGlpAction(
                market=self.market_info,
                token=token.name,
                token_amount=amount,
                mint_amount=mint_amount,
            )
        )
        return mint_amount / 10 ** self.glp_decimal

    def _remove_liquidity(self,
                          token: TokenInfo,
                          glp_amount: Decimal | float):
        glp_supply = Decimal(self.market_status.data.glp)
        aum = Decimal(self.market_status.data.aum)
        aum_in_usdg = aum / Decimal(10 ** 12)
        aum_in_usdg = aum_in_usdg.quantize(Decimal('0'), rounding=ROUND_DOWN)
        usdg_amount = glp_amount * 10 ** self.glp_decimal / glp_supply * aum_in_usdg
        usdg_amount = usdg_amount.quantize(Decimal('0'), rounding=ROUND_DOWN)
        token_out = self.sell_usdg(token, usdg_amount) / 10 ** token.decimal
        self._record_action(
            SellGlpAction(
                market=self.market_info,
                token=token.name,
                glp_amount=glp_amount,
                token_out=token_out
            )
        )
        return token_out

    def buy_usdg(self, token: TokenInfo, token_amount: Decimal | float):
        # function buyUSDG
        price = self.market_status.data[f"{token.name.lower()}_price"]
        usdg_amount = token_amount * 10 ** token.decimal * price / 10 ** 30
        usdg_amount = usdg_amount.quantize(Decimal('0'), rounding=ROUND_DOWN)
        fee_basic_point = self.get_buy_usdg_fee_point(token, usdg_amount)
        amount_after_fee = self._collect_swap_fee(token, token_amount, fee_basic_point)
        mint_amount = Decimal(amount_after_fee) * 10 ** token.decimal * price / 10 ** 30
        mint_amount = mint_amount.quantize(Decimal('0'), rounding=ROUND_DOWN)
        return mint_amount

    def sell_usdg(self, token: TokenInfo, usdg_amount: Decimal | float):
        redemption_amount = self.get_redemption_amount(token, usdg_amount)
        fee_basic_point = self.get_sell_usdg_fee_point(token, usdg_amount)
        amount_after_fee = self._collect_swap_fee(token, redemption_amount, fee_basic_point)
        return amount_after_fee

    def get_redemption_amount(self, token: TokenInfo, usdg_amount: Decimal | float):
        price = Decimal(self.market_status.data[f"{token.name.lower()}_price"]) / self.PRICE_PRECISION
        redemption_amount = usdg_amount / Decimal(price)
        return redemption_amount

    def get_buy_usdg_fee_point(self,
                               token: TokenInfo,
                               usdg_amount: Decimal | float):
        # not consider
        return self.get_fee_basis_points(token, usdg_amount, True)

    def get_sell_usdg_fee_point(self,
                                token: TokenInfo,
                                usdg_amount: Decimal | float):
        # not consider
        return self.get_fee_basis_points(token, usdg_amount, False)

    def get_fee_basis_points(self,
                             token: TokenInfo,
                             usdg_amount: Decimal | float,
                             increase: bool):
        initial_amount = Decimal(self.market_status.data[f"{token.name.lower()}_usdg"])
        next_amount = initial_amount + usdg_amount
        if not increase:
            next_amount = 0 if usdg_amount > initial_amount else initial_amount - usdg_amount
        target_amount = self.get_target_amount(token)
        if target_amount == 0:
            return self.mint_burn_fee_basis_points
        initial_diff = initial_amount - target_amount if initial_amount > target_amount else target_amount - initial_amount
        next_diff = next_amount - target_amount if next_amount > target_amount else target_amount - next_amount
        if next_diff < initial_diff:
            rebate_bps = self.tax_basis_points * initial_diff / target_amount
            return 0 if rebate_bps > self.mint_burn_fee_basis_points else self.mint_burn_fee_basis_points - rebate_bps
        average_diff = (initial_diff + next_diff) / 2
        if average_diff > target_amount:
            average_diff = target_amount
        tax_bps = self.tax_basis_points * average_diff / target_amount
        return self.mint_burn_fee_basis_points + int(tax_bps)

    def get_target_amount(self, token: TokenInfo):
        total_token_weights = 0
        for _token in self._tokens:
            total_token_weights += self.market_status.data[f"{_token.name.lower()}_weight"]
        supply = Decimal(self.market_status.data['usdg'])
        weight = self.market_status.data[f"{token.name.lower()}_weight"]
        return weight * supply / total_token_weights

    def _collect_swap_fee(self,
                          token: TokenInfo,
                          token_amount: Decimal | float,
                          fee_point: Decimal | float):
        fee_amount = token_amount * fee_point / 10000
        after_fee_amount = token_amount - fee_amount
        return after_fee_amount

    def load_data(self, chain: ChainType, start_date: date, end_date: date) -> None:
        self.logger.info('start load files from {start_date} to {end_date}...')
        assert start_date <= end_date, f'start date {start_date} should earlier than end date {end_date}'
        df = pd.DataFrame()
        day = start_date
        while day <= end_date:
            data_path = os.path.join(self.data_path, f"{chain.name.lower()}_gmx_{day.strftime('%Y-%m-%d')}.csv")
            day_df = pd.read_csv(
                data_path,
                index_col=0,
                parse_dates=True,
                converters={
                    'glp_price': to_decimal,
                    'weth_price': to_decimal,
                    'wavax_price': to_decimal,
                    'glp': to_decimal,
                    'aum': to_decimal,
                }
            )
            if len(day_df.index) > 0:
                df = pd.concat([df, day_df])
            day = day + timedelta(days=1)
        self.data = df
        self.logger.info("data has been prepared")

    def get_price_from_data(self):
        if self.data is None:
            raise DemeterError("data has not set")
        # keep weth/wavax
        df_price = self.data[['weth_price', 'wavax_price']]
        df_price['weth_price'] = df_price['weth_price'].apply(lambda r: r / self.PRICE_PRECISION)
        df_price['wavax_price'] = df_price['wavax_price'].apply(lambda r: r / self.PRICE_PRECISION)
        df_price.rename(columns={'weth_price': 'WETH', 'wavax_price': 'WAVAX'}, inplace=True)
        return df_price

    def get_market_balance(self) -> GmxBalance:
        # market 有glp + weth/avax fee  change by chain config
        net_value = self.glp_amount * Decimal(self.market_status.data.glp_price) + self.reward * Decimal(self.market_status.data['wavax_price']) / self.PRICE_PRECISION
        val = GmxBalance(
            net_value=net_value,
            reward=self.reward,
            glp=self.glp_amount
        )
        return val

    def set_market_status(  # 链上数据回测，gmx aave，跟踪
        self,
        data: MarketStatus,
        price: pd.Series,
    ):
        super().set_market_status(data, price)
        if data.data is None:
            data.data = self.data.loc[data.timestamp]
        self._market_status = data

    def check_market(self):
        super().check_market()

    def _resample(self, freq: str):
        self._data = self.data.resample(freq).first()


import numpy as np
import pandas as pd
from demeter import DemeterError


def return_value(init_equity: float, final_equity: float) -> float:
    """
    Get return rate

    :param init_equity: init equity
    :param final_equity: final equity
    :return: return value
    """
    return final_equity - init_equity


def return_rate(init_equity: float, final_equity: float) -> float:
    """
    rate of return, if init_equity is 0, will return np.inf

    :param init_equity: init equity
    :param final_equity: final equity
    :return: return rate
    """
    return final_equity / init_equity - 1 if init_equity > 0 else np.inf


def return_multiple(net_value: pd.Series) -> pd.Series:
    """
    get return multiple, value(t) / value(t-1), if there are nan/inf values, it will be set to 1

    :param net_value: list of values
    :return: return multiple
    """
    return (net_value / net_value.shift(1)).fillna(1).replace([np.inf, -np.inf], 1)


def return_rate_series(net_value: pd.Series) -> pd.Series:
    """
    calculate return rate of a net value series, (value(t) - value(t-1)) / value(t-1),
    if there are nan/inf values, it will be set to 0.

    :param net_value: list of values
    :return: return rate list
    """
    return net_value.pct_change().fillna(0).replace([np.inf, -np.inf], 0)


def annualized_return(
    duration_in_day: float,
    init_value: float = None,
    final_value: float = None,
    return_rates: pd.Series = None,
    net_values: pd.Series = None,
    interest_type="compound",
) -> float:
    """
    | Annualizing an asset's return rate. you can choose one way to calculate
    | * initial value and final value
    | * return rate list
    | * net value list

    :param duration_in_day: days between initial time and final time
    :param init_value: initial value
    :param final_value: final value
    :param return_rates: a list of return rate
    :param net_values: a list of net value
    :param interest_type: interest type, can be "single" or "compound", default is compound
    :return: annualized return value:
    """

    if interest_type == "single":
        if init_value is not None and final_value is not None:
            return ((final_value - init_value) / init_value) / (duration_in_day / 365)
        elif net_values is not None:
            return (net_values.iloc[-1] - net_values.iloc[0]) / net_values.iloc[0] / (duration_in_day / 365)
        elif return_rates is not None:
            raise DemeterError("single interest series should not exist, it should be compound")
        else:
            raise DemeterError("Choose one to calculate: initial and final value/return rate/net value")
    elif interest_type == "compound":
        if init_value is not None and final_value is not None:
            return (final_value / init_value) ** (365 / duration_in_day) - 1
        elif net_values is not None:
            return_rates = return_multiple(net_values)
            return return_rates.prod() ** (365 / duration_in_day) - 1
        elif return_rates is not None:
            return (return_rates + 1).prod() ** (365 / duration_in_day) - 1
        else:
            raise DemeterError("Choose one to calculate: initial and final value/return rate/net value")


def max_draw_down_benchmark(value: pd.Series):
    """
    Calculate the maximum drawdown based on its definition.
    This method is computationally intensive and is used for comparing results with max_draw_down.

    :param value: value to calculate
    :type value: pd.Series
    """
    value.index = range(len(value.index))  # restruct index to access faster
    result = 0
    for index, row in value.iteritems():
        current_max = value[index:].apply(lambda nv: 1 - nv / row).max()
        if current_max > result:
            result = current_max
    return result


def max_draw_down(net_value: pd.Series):
    """
    Get max draw down in a fast algorithm.

    :param net_value: value to calculate
    :type net_value:  pd.Series
    """
    max_value, idx_h, idx_l = _withdraw_with_high_low(net_value.to_list())
    return (net_value.iloc[idx_h] - net_value.iloc[idx_l]) / net_value.iloc[idx_h]


def _withdraw_with_high_low(arr: list):
    """
    from : https://blog.csdn.net/Spade_/article/details/112341428
    """

    # Given an array, return the maximum drawdown and the corresponding indices of the highest and lowest points.
    _dp = 0  # Let _dp[i] represent the maximum drawdown ending at index i.
    i_high = 0  # During iteration, the index of the highest point within the range from 0 to i - 1.

    # Global maximum drawdown and the corresponding highest and lowest point indices, noting that indices are meant.
    g_withdraw, g_high, g_low = -np.inf, -1, -1

    for i in range(1, len(arr)):
        # If the highest point within the range from 0 to i - 1 is less than the current point...
        if arr[i_high] < arr[i - 1]:
            i_high = i - 1  # ...the index of the highest point within the range from 0 to i - 1
        _dp = arr[i_high] - arr[i]  # ...then _dp[i] is updated to represent the maximum drawdown ending at index i.
        if _dp > g_withdraw:  # Find the new maximum drawdown and update the three values:
            g_withdraw = _dp
            g_high = i_high
            g_low = i

    return g_withdraw, g_high, g_low


def volatility(returns: pd.Series, interval_in_day):
    """
    Calculate volatility, The number of trading days is 365 instead of 252.

    :param interval_in_day: Daily data interval.
    :param returns: list of values
    :return: volatility value
    """
    return returns.std() * np.sqrt(365 / interval_in_day)


def sharpe_ratio(interval_in_day: float, duration_in_day: int, values: pd.Series, annualized_risk_free_rate: float):
    """
    Calculate sharpe ratio. The number of trading days is 365 instead of 252.

    :param interval_in_day: Daily data interval.
    :param duration_in_day: days between initial time and final time
    :param values: list of values
    :param annualized_risk_free_rate: annualized risk free rate
    :return: sharpe ratio
    """

    returns = (values / values.shift(1)).dropna()

    mean_yearly_return = annualized_return(duration_in_day, return_rates=returns - 1)
    std_yearly_return = volatility(returns, interval_in_day)

    result = (mean_yearly_return - annualized_risk_free_rate) / std_yearly_return

    return result


def alpha_beta(values: pd.Series, benchmark: pd.Series, duration_in_day: int):
    """
    Calculate alpha and beta

    :param values: list of values
    :param benchmark: benchmark
    :param duration_in_day: days between initial time and final time

    :return: alpha and beta
    """
    portfolio_returns = (values / values.shift(1)).dropna()
    benchmark_returns = (benchmark / benchmark.shift(1)).dropna()

    cov = np.cov(portfolio_returns, benchmark_returns)
    beta = cov[0, 1] / cov[1, 1]

    portfolio_apy = annualized_return(duration_in_day, return_rates=portfolio_returns - 1)
    benchmark_apy = annualized_return(duration_in_day, return_rates=benchmark_returns - 1)

    alpha = portfolio_apy - beta * benchmark_apy

    return alpha, beta



from typing import Dict, Any

from decimal import Decimal

from ._typing import MetricEnum
from .calculator import *

DECIMAL_1 = Decimal(1)


def performance_metrics(
    values: pd.Series, annualized_risk_free_rate=0.03, benchmark: pd.Series | None = None
) -> Dict[MetricEnum, Any]:
    """
    Calculate all performance metrics

    :param values: value's you need to calculate,
    :param annualized_risk_free_rate: annualized risk_free rate
    :param benchmark: benchmark, if set to None, some metrics depends on this will not be calculated
    :return: a dict with metric enum and their value.
    """
    values = values.apply(lambda x: float(x))
    init = values.iloc[0]
    final = values.iloc[-1]

    start = values.index[0]
    start1 = values.index[1]
    interval = start1 - start
    interval_in_day = interval.value / 1e9 / 86400
    end = values.index[len(values) - 1]
    duration_in_day = (end - start + interval).value / 1e9 / 86400

    if benchmark is not None:
        benchmark = benchmark.apply(lambda x: float(x))
        alpha, beta = alpha_beta(values, benchmark, duration_in_day)
        benchmark_init = benchmark.iloc[0]
        benchmark_final = benchmark.iloc[-1]
        benchmark_return = return_rate(benchmark_init, benchmark_final)
        benchmark_apr = annualized_return(duration_in_day, benchmark_init, benchmark_final)
    else:
        alpha, beta, benchmark_return, benchmark_apr = np.nan, np.nan, np.nan, np.nan

    returns = values.pct_change().dropna()
    metric_map = {
        MetricEnum.start_period: values.index[0],
        MetricEnum.end_period: values.index[-1],
        MetricEnum.start_val: init,
        MetricEnum.end_val: final,
        MetricEnum.duration: (values.index[-1] - values.index[0]) + interval,
        MetricEnum.return_value: return_value(init, final),
        MetricEnum.return_rate: return_rate(init, final),
        MetricEnum.annualized_return: annualized_return(duration_in_day, init, final),
        MetricEnum.max_draw_down: max_draw_down(values),
        MetricEnum.sharpe_ratio: sharpe_ratio(interval_in_day, duration_in_day, values, annualized_risk_free_rate),
        MetricEnum.volatility: volatility(returns, interval_in_day),
        MetricEnum.alpha: alpha,
        MetricEnum.beta: beta,
        MetricEnum.benchmark_rate: benchmark_return,
        MetricEnum.annualized_benchmark_rate: benchmark_apr,
    }
    return {k: v for k, v in metric_map.items()}


def round_results(val_dict: Dict[MetricEnum, Any], decimal: int = 3):
    return {k: round(v, decimal) if isinstance(v, (int, float, complex, Decimal)) else v for k, v in val_dict.items()}



"""
Indicator is a group of calculators, can be used in strategy or somewhere else.
"""


from .ma import simple_moving_average, exponential_moving_average
from .volatility import realized_volatility



"""
Aave market, this module can simulate some operations in Aave, such as supply/withdraw/borrow/repay

but flash debt is not supported.
Liquidate is not supported too, but your debt will be liquidated if health factor is too low.
"""

from ._typing import (
    AaveTokenStatus,
    SupplyInfo,
    BorrowInfo,
    InterestRateMode,
    SupplyKey,
    BorrowKey,
    AaveBalance,
    Supply,
    Borrow,
    RiskParameter,
    LiquidationAction,
    RepayAction,
    BorrowAction,
    WithdrawAction,
    SupplyAction,
    AaveMarketStatus,
    AaveDescription,
)
from .core import AaveV3CoreLib
from .market import AaveV3Market



import logging
from decimal import Decimal
from functools import wraps
from typing import Dict, Callable

import pandas as pd

from ._typing import BaseAction, MarketBalance, MarketStatus, MarketInfo, RowData
from .._typing import DECIMAL_0, DemeterError, TokenInfo, USD

DEFAULT_DATA_PATH = "./data"


def write_func(func):
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        instance = args[0]
        if not instance.is_open:
            raise DemeterError(f"{instance.market_info.name} is not open.")
        ret = func(*args, **kwargs)
        instance.has_update = True
        return ret

    return wrapper_func


class Market:
    """

    | Market is the place to invest your assets.
    | This is an abstract class, you should use subclass instead this one

    :param market_info: Key of this market.
    :type market_info: MarketInfo
    :param data: Data is used to simulate market status. usually it is downloaded from log event of ethereum, then indexed by timestamp with minute interval.
    :type data: DataFrame
    :param data_path: default folder path for data files. Each day has a corresponding csv file.
    :type data_path: str
    """

    def __init__(self, market_info: MarketInfo, data: pd.DataFrame = None, data_path=DEFAULT_DATA_PATH):
        """
        Initialize a Market

        """
        self._data: pd.DataFrame = data
        self._market_info: MarketInfo = market_info
        self.broker = None
        self._record_action_callback: Callable[[BaseAction], None] = None
        self.data_path: str = data_path
        self.logger = logging.getLogger(__name__)
        self._market_status: MarketStatus = MarketStatus(None, pd.Series())
        self._price_status: pd.Series | None = None
        # if some var that related to market status has changed, should set this to True,
        # then the second set_market_status in every minute will be triggerd
        # e.g. At uniswap market, when add liquidity in the head of the minute, this flag will be set to True,
        # so user liquidity will be added to total liquidity in this minute, and get more fee
        # remember set this flag to False after set_market_status
        self.has_update = False
        self.open: Callable[[RowData], None] = None
        # if market interval is minutely, is_open will always true,
        # or it will be false until timestamp is on its interval
        self.is_open: bool = True
        self.quote_token: TokenInfo = USD

    def __str__(self):
        return f"{self._market_info.name}:{type(self).__name__}"

    @property
    def market_info(self) -> MarketInfo:
        """
        Get market info, it is the key of a market.
        """
        return self._market_info

    @property
    def data(self) -> pd.DataFrame:
        """
        | Market data is used to simulate the status of market. For example, in uniswap market, data describe pool liquidity, price of this pool.
        | Usually data is got by demeter-fetch which can download and decode on chain event log. Data will be saved in CSV format, each day has a corresponding csv file.
        | Those csv file is indexed by timestamp, and resampled to one minute.
        | Data files will be loaded as dataframe. The whole back test process is based on minutely timestamp.

        :return: market data
        :rtype: DataFrame
        """
        return self._data

    @data.setter
    def data(self, value):
        """
        Set data and check its type
        """
        if isinstance(value, pd.DataFrame):
            self._data = value
        else:
            raise ValueError()

    def _record_action(self, action: BaseAction):
        if self._record_action_callback is not None:
            self._record_action_callback(action)

    # region for subclass to override
    def check_market(self):
        """
        check market before back test
        """
        if not isinstance(self.data, pd.DataFrame):
            raise DemeterError("data must be type of data frame")
        if not isinstance(self.data.index, pd.core.indexes.datetimes.DatetimeIndex):
            raise DemeterError("date index must be datetime")

    def update(self):
        """
        Update market in every loop. It was used for triggering market status calculation.
        """
        pass

    @property
    def market_status(self):
        """
        Get market status, such as current total liquidity, current apy, etc.
        In short, it's a row of market.data.
        """
        return self._market_status

    def set_market_status(
        self,
        data: MarketStatus,
        price: pd.Series,
    ):
        """
        Set up market status, such as liquidity, price

        :param data: market status
        :type data: Series | MarketStatus
        :param price: current price
        :type price: Series

        """
        # self._market_status = data
        self._price_status = price
        self.is_open = True if self._data is None or data.timestamp in self._data.index else False
        self.has_update = False

    def get_market_balance(self) -> MarketBalance:
        """
        Get market asset balance, such as current positions, net values

        :return: Balance in this market includes net value, position value
        :rtype: MarketBalance
        """
        return MarketBalance(DECIMAL_0)

    def formatted_str(self):
        """
        Get a colorful brief description to print in console.
        """
        return ""

    def _resample(self, freq: str):
        """
        Resample data in this market
        """
        pass

    # endregion



from dataclasses import dataclass, field
from datetime import datetime, date
from decimal import Decimal
from typing import List

from demeter import TokenInfo, AccountStatus, BaseAction, MarketInfo
from demeter._typing import MarketDescription, DemeterLog


@dataclass
class BackTestDescription:
    strategy_name: str
    quote_token: TokenInfo
    init_status: AccountStatus
    assets: List[TokenInfo]
    markets: List[MarketDescription]
    actions: List[BaseAction]
    backtest_start: datetime
    backtest_end: datetime
    backtest_duration: float
    logs: List[DemeterLog] = field(default_factory=list)


@dataclass
class Position:
    key: any
    market: MarketInfo
    start: datetime
    end: datetime | None
    amount: Decimal


@dataclass
class OptionPosition(Position):
    token: str
    strike_price: int
    expiry_time: datetime
    type: str  # call put


@dataclass
class LpPosition(Position):
    base_token: str
    quote_token: str
    lower_price: Decimal
    upper_price: Decimal



"""
This package includes modules relate to strategy,
"""

from .strategy import Strategy
from .trigger import (
    Trigger,
    TimeRangesTrigger,
    TimeRangeTrigger,
    TimeRange,
    PeriodTrigger,
    PeriodsTrigger,
    AtTimesTrigger,
    AtTimeTrigger,
    PriceTrigger,
)



from decimal import Decimal
from enum import Enum
from typing import Dict

import pandas as pd

from demeter import Formats, UnitDecimal


class ForColorEnum(Enum):
    """
    foreground color in console
    """

    default = 0
    black = 30
    red = 31
    green = 32
    yellow = 33
    blue = 34
    purple = 35
    cyan = 36
    white = 37
    dark_grey = 90
    light_red = 91
    light_green = 92
    light_yellow = 93
    light_blue = 94
    pink = 95
    light_cyan = 96


class BackColorEnum(Enum):
    """
    background color in console
    """

    default = 0
    black = 40
    red = 41
    green = 42
    yellow = 43
    blue = 44
    purple = 45
    cyan = 46
    white = 47


class ModeEnum(Enum):
    """
    font mode in console
    """

    normal = 0
    bold = 1
    underline = 4
    blink = 5
    invert = 7
    hide = 8


DEFAULT_END = 0

STYLE = {
    "header1": {
        "mode": ModeEnum.invert,
        "fore": ForColorEnum.red,
        "back": BackColorEnum.default,
        "width": 50,
    },
    "header2": {
        "mode": ModeEnum.invert,
        "fore": ForColorEnum.purple,
        "back": BackColorEnum.default,
        "width": 30,
    },
    "header3": {
        "mode": ModeEnum.underline,
        "fore": ForColorEnum.yellow,
        "back": BackColorEnum.default,
        "width": -1,
    },
    "key": {
        "mode": ModeEnum.normal,
        "fore": ForColorEnum.blue,
        "back": BackColorEnum.default,
        "width": 10,
    },
    "value": {
        "mode": ModeEnum.normal,
        "fore": ForColorEnum.default,
        "back": BackColorEnum.default,
        "width": 25,
    },
}


def get_formatted(
    string: str,
    mode: ModeEnum = ModeEnum.normal,
    fore: ForColorEnum = ForColorEnum.default,
    back: BackColorEnum = BackColorEnum.default,
    width=-1,
) -> str:
    """
    Get formatted string to display in console

    :param string: text to convert
    :type string: str
    :param mode: text mode
    :type mode: ModeEnum
    :param fore: forground color
    :type fore: ForColorEnum
    :param back: background color
    :type back: BackColorEnum
    :param width: text width
    :type width: int
    :return: string with console format
    :rtype: str
    """
    mode = "{}".format(mode.value if mode != mode.normal else "")
    fore = "{}".format(fore.value if fore != ForColorEnum.default else "")
    back = "{}".format(back.value if back != BackColorEnum.default else "")
    style = ";".join([s for s in [mode, fore, back] if s])
    end = ""
    if style != "":
        style = """\033[{}m""".format(style)
        end = """\033[0m"""
    if width > 0:
        string = "{}{:<{}}{}".format(style, string, width, end)
    else:
        string = "{}{}{}".format(style, string, end)

    return string


def get_formatted_predefined(string: str, style: Dict) -> str:
    """
    Get formatted string with predefined rule

    :param string: text to convert
    :type string: str
    :param style: style defined in STYLE dictionary
    :type style: Dict
    :return: string with console format
    :rtype: str
    """
    return get_formatted(string, style["mode"], style["fore"], style["back"], style["width"])


def get_formatted_from_dict(values: Dict[str, str]) -> str:
    """
    Get formatted string of key:value pair with predefined rule

    :param values: dict to convert, e.g. name: Eric.
    :type values: Dict[str, str]
    :return: string with console format
    :rtype: str
    """
    str_array = []
    for k, v in values.items():
        if isinstance(v, (float, Decimal)):
            v_str = "{num:{f}}".format(num=v, f=Formats.global_num_format)
        else:
            v_str = v
        str_array.append(
            f"{get_formatted_predefined(k, STYLE['key'])}:{get_formatted_predefined(v_str, STYLE['value'])}"
        )
    return "".join(str_array)


def get_action_str(action, title_color: ForColorEnum, value_dict: Dict[str, str]) -> str:
    title = f"""\n\033[1;42m{action.timestamp.strftime("%Y-%m-%d %H:%M:%S")}\033[0m \033[1;95m{action.market.name:<20}\033[0m \033[1;{title_color.value}m{action.action_type.value:<20}\033[0m"""
    body = get_formatted_from_dict(value_dict)
    if action.comment is not None:
        body += "\t\033[4;35m" + action.comment + "\033[0m"
    return title + body


def print_dataframe_with_precision(df):
    format_str = "{:" + Formats.global_num_format + "}"
    # df_float = df.apply(lambda x: pd.to_numeric(x, downcast="float") if pd.api.types.is_numeric_dtype(type(x)) else x)
    with pd.option_context("display.float_format", format_str.format):
        print(df.astype(float))


def format_value(value) -> str:
    if isinstance(value, UnitDecimal):
        return value.to_str()
    elif isinstance(value, (float, Decimal)):
        return "{num:{f}}".format(num=value, f=Formats.global_num_format)
    elif isinstance(value, Enum):
        return value.name
    return str(value)



from typing import List, Callable

import pandas as pd

from .trigger import Trigger
from .. import Broker, MarketDict, AccountStatus, AssetDict, Asset, RowData
from .._typing import DemeterError
from ..broker import MarketInfo, BaseAction, Market


class Strategy(object):
    """
    Parent class of strategy, all user's strategy should inherit this class
    """

    def __init__(self):
        self.broker: Broker | None = None
        self.data: MarketDict[pd.DataFrame] = MarketDict()
        self.markets: MarketDict[Market] = MarketDict()
        self.prices: pd.DataFrame | None = None
        self.triggers: [Trigger] = []
        self.account_status: List[AccountStatus] = []
        self.account_status_df: pd.DataFrame | None = None
        self.comment_last_action: Callable = lambda msg: msg
        self.assets: AssetDict[Asset] = AssetDict()
        self.actions: List[BaseAction] = []
        self.log: Callable = lambda t, msg, info: msg

    def initialize(self):
        """
        Initialize your strategy, this will be called before iteration start

        """
        pass

    def on_bar(self, row_data: RowData):
        """
        Called after triggers on each iteration, at this time, market are not updated yet(Take uniswap market for example, fee of this minute are not added to positions).

        :param row_data: data in this iteration, include current timestamp, price, all columns data, and indicators(such as simple moving average)
        :type row_data: RowData
        """
        pass

    def after_bar(self, row_data: RowData):
        """
        called after market are updated on each iteration

        :param row_data: data in this iteration, include current timestamp, price, all columns data, and indicators(such as simple moving average)
        :type row_data: RowData
        """
        pass

    def finalize(self):
        """
        this will run after all the data processed. You can access broker.account_status, broker.market.status to do some calculation

        """
        pass

    def notify(self, action: BaseAction):
        """
        Notify when an action(buy/sell) happens

        :param action:  action taken in market.
        :type action: BaseAction
        """
        pass

    def add_column(self, market: MarketInfo | Market, name: str, column_data: pd.Series):
        """
        add a column to data in a market

        :param name: column name, like sma
        :type name: str
        :param market: which market to update
        :type market:  MarketInfo | Market
        :param column_data: One data column, it should have the same timestamp index with market.data
        :type column_data: pd.Series
        """
        if not isinstance(column_data.index, pd.core.indexes.datetimes.DatetimeIndex):
            raise DemeterError("date index must be datetime")
        if isinstance(market, MarketInfo):
            self.broker.markets[market].data[name] = column_data
        elif isinstance(market, Market):
            market.data[name] = column_data
        else:
            raise DemeterError(f"{market} is not a valid market")



from enum import Enum


class MetricEnum(Enum):
    start_period = "Start period"
    end_period = "End period"
    start_val = "Start Value"
    end_val = "End Value"
    duration = "Duration"
    return_value = "Return"
    return_rate = "Rate of Return"
    annualized_return = "APR"
    max_draw_down = "Max Draw Down"
    sharpe_ratio = "Sharpe Ratio"
    volatility = "Volatility"
    alpha = "Alpha"
    beta = "Beta"
    benchmark_rate = "Benchmark return rate"
    annualized_benchmark_rate = "Benchmark APR"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

